/* giramobject.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999 David Odin <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <gtk/gtk.h>
#include "vectors.h"
#include "giramlibprivate.h"
#include "protocol.h"
#include "giramlib.h"

/*****************************************************************************
*  giramlib_box_create
******************************************************************************/
int giramlib_box_create(Vector Vect1, Vector Vect2)
{
  GiramObject *Box;
  GiramBox    *BBox;

  Box = g_malloc(sizeof(GiramBox));
  BBox = (GiramBox *)Box;
  Box->Next = GiramFirstObject;
  GiramFirstObject = Box;
  Box->Type = BoxType;
  Box->Realised = FALSE;
  Box->Id = NextObjectId++;
  Box->Trans = NULL;
  BBox->MinCorner[0] = Vect1[0];
  BBox->MinCorner[1] = Vect1[1];
  BBox->MinCorner[2] = Vect1[2];
  BBox->MaxCorner[0] = Vect2[0];
  BBox->MaxCorner[1] = Vect2[1];
  BBox->MaxCorner[2] = Vect2[2];
  return Box->Id;
}

/*****************************************************************************
*  giram_cone_create
******************************************************************************/
int giramlib_cone_create(Vector Base, double BaseRadius, 
                         Vector Apex, double ApexRadius)
{
  GiramObject *Cone;
  GiramCone   *CCone;

  Cone = g_malloc(sizeof(GiramCone));
  CCone = (GiramCone *)Cone;
  Cone->Next = GiramFirstObject;
  GiramFirstObject = Cone;
  Cone->Type = ConeType;
  Cone->Realised = FALSE;
  Cone->Id = NextObjectId++;
  Cone->Trans = NULL;
  CCone->Base[0] = Base[0];
  CCone->Base[1] = Base[1];
  CCone->Base[2] = Base[2];
  CCone->BaseRadius = BaseRadius;
  CCone->Apex[0] = Apex[0];
  CCone->Apex[1] = Apex[1];
  CCone->Apex[2] = Apex[2];
  CCone->ApexRadius = ApexRadius;
  return Cone->Id;
}

/*****************************************************************************
*  giram_cylinder_create
******************************************************************************/
int giramlib_cylinder_create(Vector Base, Vector Apex, double Radius)
{
  GiramObject   *Cylinder;
  GiramCylinder *CCylinder;

  Cylinder = g_malloc(sizeof(GiramCylinder));
  CCylinder = (GiramCylinder *)Cylinder;
  Cylinder->Next = GiramFirstObject;
  GiramFirstObject = Cylinder;
  Cylinder->Type = ConeType;
  Cylinder->Realised = TRUE;
  Cylinder->Id = NextObjectId++;
  Cylinder->Trans = NULL;
  CCylinder->Base[0] = Base[0];
  CCylinder->Base[1] = Base[1];
  CCylinder->Base[2] = Base[2];
  CCylinder->Apex[0] = Apex[0];
  CCylinder->Apex[1] = Apex[1];
  CCylinder->Apex[2] = Apex[2];
  CCylinder->Radius = Radius;
  return Cylinder->Id;
}

/*****************************************************************************
*  giram_sphere_create
******************************************************************************/
int giramlib_sphere_create(Vector Center, double Radius)
{
  GiramObject *Sphere;
  GiramSphere *SSphere;

  Sphere = g_malloc(sizeof(GiramSphere));
  SSphere = (GiramSphere *)Sphere;
  Sphere->Next = GiramFirstObject;
  GiramFirstObject = Sphere;
  Sphere->Type = SphereType;
  Sphere->Realised = FALSE;
  Sphere->Id = NextObjectId++;
  Sphere->Trans = NULL;
  SSphere->Center[0] = Center[0];
  SSphere->Center[1] = Center[1];
  SSphere->Center[2] = Center[2];
  SSphere->Radius = Radius;
  return Sphere->Id;
}

/*****************************************************************************
*  giram_triangle_create
******************************************************************************/
int giramlib_triangle_create(Vector V1, Vector V2, Vector V3)
{
  GiramObject   *Triangle;
  GiramTriangle *TTriangle;

  Triangle = g_malloc(sizeof(GiramTriangle));
  TTriangle = (GiramTriangle *)Triangle;
  Triangle->Next = GiramFirstObject;
  GiramFirstObject = Triangle;
  Triangle->Type = TriangleType;
  Triangle->Realised = FALSE;
  Triangle->Id = NextObjectId++;
  Triangle->Trans = NULL;
  TTriangle->P1[0] = V1[0];
  TTriangle->P1[1] = V1[1];
  TTriangle->P1[2] = V1[2];
  TTriangle->P2[0] = V2[0];
  TTriangle->P2[1] = V2[1];
  TTriangle->P2[2] = V2[2];
  TTriangle->P3[0] = V3[0];
  TTriangle->P3[1] = V3[1];
  TTriangle->P3[2] = V3[2];
  return Triangle->Id;
}

/*****************************************************************************
*  giram_object_realize
******************************************************************************/
void giramlib_object_realize(int ObjectId)
{
  GiramObject *TmpObject;
  int ObjectID = -1;

  for (TmpObject = GiramFirstObject;
       TmpObject && TmpObject->Id != ObjectId;
       TmpObject = TmpObject->Next)
    ;
  if (TmpObject == NULL || TmpObject->Realised)
    return;
  switch (TmpObject->Type)
  {
    case BicubicPatchType:

      break;
    case BlobType:

      break;
    case BoxType:
      {
        GiramBox *BBox = (GiramBox *)TmpObject;

        ObjectID = giram_send_box(BBox->MinCorner, BBox->MaxCorner);
      }
      break;
    case ConeType:
      {
        GiramCone *CCone = (GiramCone *)TmpObject;

        ObjectID = giram_send_cone(CCone->Base, CCone->BaseRadius,
                                   CCone->Apex, CCone->ApexRadius);
      }
      break;
    case CSGType:

      break;
    case CubicType:

      break;
    case CylinderType:
      {
        GiramCylinder *CCylinder = (GiramCylinder *)TmpObject;

        ObjectID = giram_send_cylinder(CCylinder->Base, CCylinder->Apex,
                                       CCylinder->Radius);
      }
      break;
    case DiscType:
      {
        GiramDisc *DDisc = (GiramDisc *)TmpObject;

        ObjectID = giram_send_disc(DDisc->Center, DDisc->Normal,
                                   DDisc->Radius, DDisc->HoleRadius);
      }
      break;
    case HeightFieldType:

      break;
    case JuliaFractalType:

      break;
    case LatheType:

      break;
    case MeshType:

      break;
    case PlaneType:
      {
        GiramPlane *PPlane = (GiramPlane *)TmpObject;

        ObjectID = giram_send_plane(PPlane->Normal, PPlane->Distance);
      }
      break;
    case PolyType:

      break;
    case PolygonType:

      break;
    case PrismType:

      break;
    case QuadricType:

      break;
    case QuarticType:

      break;
    case SorType:

      break;
    case SphereType:
      {
        GiramSphere *SSphere = (GiramSphere *)TmpObject;

        ObjectID = giram_send_sphere(SSphere->Center, SSphere->Radius);
      }
      break;
    case SuperEllipsoidType:
      {
        GiramSuperEllipsoid *SSuperEllipsoid = (GiramSuperEllipsoid *)TmpObject;

        ObjectID = giram_send_superellipsoid(SSuperEllipsoid->E, SSuperEllipsoid->N);
      }
      break;
    case TextType:

      break;
    case TorusType:
      {
        GiramTorus *TTorus = (GiramTorus *)TmpObject;

        ObjectID = giram_send_torus(TTorus->Major, TTorus->Minor);
      }
      break;
    case TriangleType:
      {
        GiramTriangle *TTriangle = (GiramTriangle *)TmpObject;

        ObjectID = giram_send_triangle(TTriangle->P1, TTriangle->P2, TTriangle->P3);
      }
      break;
  }
  if (TmpObject->Trans)
  {
    giram_send_trans(ObjectID, TmpObject->Trans);
  }
  TmpObject->Realised = TRUE;
}

/*****************************************************************************
*  giram_object_realize_all
******************************************************************************/
void giramlib_object_realize_all(void)
{
  GiramObject *TmpObject;
  int ObjectID = -1;

  for (TmpObject = GiramFirstObject; TmpObject ; TmpObject = TmpObject->Next)
  {
    if (!(TmpObject->Realised))
    {
      switch (TmpObject->Type)
      {
        case BicubicPatchType:

          break;
        case BlobType:

          break;
        case BoxType:
          {
            GiramBox *BBox = (GiramBox *)TmpObject;

            ObjectID = giram_send_box(BBox->MinCorner, BBox->MaxCorner);
          }
          break;
        case ConeType:
          {
            GiramCone *CCone = (GiramCone *)TmpObject;

            ObjectID = giram_send_cone(CCone->Base, CCone->BaseRadius,
                                       CCone->Apex, CCone->ApexRadius);
          }
          break;
        case CSGType:

          break;
        case CubicType:

          break;
        case CylinderType:
          {
            GiramCylinder *CCylinder = (GiramCylinder *)TmpObject;

            ObjectID = giram_send_cylinder(CCylinder->Base, CCylinder->Apex,
                                           CCylinder->Radius);
          }
          break;
        case DiscType:
          {
            GiramDisc *DDisc = (GiramDisc *)TmpObject;

            ObjectID = giram_send_disc(DDisc->Center, DDisc->Normal,
                                       DDisc->Radius, DDisc->HoleRadius);
          }
          break;
        case HeightFieldType:

          break;
        case JuliaFractalType:

          break;
        case LatheType:

          break;
        case MeshType:

          break;
        case PlaneType:
          {
            GiramPlane *PPlane = (GiramPlane *)TmpObject;

            ObjectID = giram_send_plane(PPlane->Normal, PPlane->Distance);
          }
          break;
        case PolyType:

          break;
        case PolygonType:

          break;
        case PrismType:

          break;
        case QuadricType:

          break;
        case QuarticType:

          break;
        case SorType:

          break;
        case SphereType:
          {
            GiramSphere *SSphere = (GiramSphere *)TmpObject;

            ObjectID = giram_send_sphere(SSphere->Center, SSphere->Radius);
          }
          break;
        case SuperEllipsoidType:
          {
            GiramSuperEllipsoid *SSuperEllipsoid = (GiramSuperEllipsoid *)TmpObject;

            ObjectID = giram_send_superellipsoid(SSuperEllipsoid->E,
                                                 SSuperEllipsoid->N);
          }
          break;
        case TextType:

          break;
        case TorusType:
          {
            GiramTorus *TTorus = (GiramTorus *)TmpObject;

            ObjectID = giram_send_torus(TTorus->Major, TTorus->Minor);
          }
          break;
        case TriangleType:
          {
            GiramTriangle *TTriangle = (GiramTriangle *)TmpObject;

            ObjectID = giram_send_triangle(TTriangle->P1, TTriangle->P2,
                                           TTriangle->P3);
          }
          break;
      }
      if (TmpObject->Trans)
      {
        giram_send_trans(ObjectID, TmpObject->Trans);
      }
      TmpObject->Realised = TRUE;
    }
  }
}

/*****************************************************************************
*  giram_object_translate
******************************************************************************/
int giramlib_object_translate(int objectID, Vector TransVect)
{
  GiramObject *TmpObject;
  TransformStruct Trans;

  for (TmpObject = GiramFirstObject;
       TmpObject && TmpObject->Id != objectID ;
       TmpObject = TmpObject->Next)
    ;
  if (TmpObject)
  {
    if (TmpObject->Trans)
    {
      ComputeTranslateTrans(&Trans, TransVect);
      ComposeTrans(TmpObject->Trans, &Trans);
    } else
    {
      TmpObject->Trans = g_malloc(sizeof(TransformStruct));
      ComputeTranslateTrans(TmpObject->Trans, TransVect);
    }
    return objectID;
  }
  return -1;
}

/*****************************************************************************
*  giram_object_rotate
******************************************************************************/
int giramlib_object_rotate(int objectID, Vector RotateVect)
{
  GiramObject *TmpObject;
  TransformStruct Trans;

  for (TmpObject = GiramFirstObject;
       TmpObject && TmpObject->Id != objectID ;
       TmpObject = TmpObject->Next)
    ;
  if (TmpObject)
  {
    if (TmpObject->Trans)
    {
      ComputeRotateTrans(&Trans, RotateVect);
      ComposeTrans(TmpObject->Trans, &Trans);
    } else
    {
      TmpObject->Trans = g_malloc(sizeof(TransformStruct));
      ComputeRotateTrans(TmpObject->Trans, RotateVect);
    }
    return objectID;
  }
  return -1;
}

/*****************************************************************************
*  giram_object_scale
******************************************************************************/
int giramlib_object_scale(int objectID, Vector ScaleVect)
{
  GiramObject *TmpObject;
  TransformStruct Trans;

  for (TmpObject = GiramFirstObject;
       TmpObject && TmpObject->Id != objectID ;
       TmpObject = TmpObject->Next)
    ;
  if (TmpObject)
  {
    if (TmpObject->Trans)
    {
      ComputeScaleTrans(&Trans, ScaleVect);
      ComposeTrans(TmpObject->Trans, &Trans);
    } else
    {
      TmpObject->Trans = g_malloc(sizeof(TransformStruct));
      ComputeScaleTrans(TmpObject->Trans, ScaleVect);
    }
    return objectID;
  }
  return -1;
}

