/* cylinder.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "giram.h"
#include "trimesh.h"
#include "csgtree.h"
#include "object.h"
#include "snap.h"
#include "utils.h"
#include "cylinder.h"

#include "giramintl.h"

int CylinderOpenFlag = 0;

static void giram_cylinder_build_triangle_mesh(ObjectStruct *cylinder);
static gboolean giram_cylinder_inside(ObjectStruct *cylinder, double x, double y, double z);
static gboolean giram_cylinder_is_intersection(ObjectStruct *cylinder, Vector org, Vector dir);
static gboolean giram_cylinder_click_in_XY(ObjectStruct *cylinder, double x, double y);
static gboolean giram_cylinder_click_in_XZ(ObjectStruct *cylinder, double x, double z);
static gboolean giram_cylinder_click_in_ZY(ObjectStruct *cylinder, double z, double y);
static gboolean giram_cylinder_find_intersection_segment(ObjectStruct *cylinder,
                                                         Vector in_point, Vector out_point,
                                                         Vector inter_point, Vector inter_norm);

/*****************************************************************************
*  giram_cylinder_build_triangle_mesh
******************************************************************************/
static void giram_cylinder_build_triangle_mesh(ObjectStruct *cylinder)
{ /* XXX: clipped_by */
  Vector P1, P2, P3, P4;
  Vector N1, N2, N3, N4;
  Vector NO1, NO2;
  Vector PO1, PO2;
  int DetailLevel;
  int i;
  CylinderStruct *ccylinder = (CylinderStruct *)cylinder;

  DetailLevel = 16; /* XXX: Should be a private field */
  
  if (cylinder->FirstTriangle)
    DestroyObjectTriangleMesh(cylinder);
  
  NO1[0] = 0.0; NO1[1] = -1.0; NO1[2] = 0.0;
  NO2[0] = 0.0; NO2[1] =  1.0; NO2[2] = 0.0;
  PO1[0] = 0.0; PO1[1] =  0.0; PO1[2] = 0.0;
  PO2[0] = 0.0; PO2[1] =  1.0; PO2[2] = 0.0;
  if (cylinder->Trans)
  {
    MEvaluatePoint(PO1, cylinder->Trans, PO1);
    MEvaluateVector(NO1, cylinder->Trans, NO1);
    MEvaluatePoint(PO2, cylinder->Trans, PO2);
    MEvaluateVector(NO2, cylinder->Trans, NO2);
  }
  for (i=0 ; i<DetailLevel ; i++)
  {
    N1[0] = P1[0] = cos((i*2.0)*M_PI/DetailLevel);
    N1[1] = P1[1] = 0.0;
    N1[2] = P1[2] = sin((i*2.0)*M_PI/DetailLevel);
    N2[0] = P2[0] = cos((i*2.0+1.0)*M_PI/DetailLevel);
    N2[1] = 0.0; P2[1] = 1.0;
    N2[2] = P2[2] = sin((i*2.0+1.0)*M_PI/DetailLevel);
    N3[0] = P3[0] = cos((i+1)*2.0*M_PI/DetailLevel);
    N3[1] = P3[1] = 0.0;
    N3[2] = P3[2] = sin((i+1)*2.0*M_PI/DetailLevel);
    N4[0] = P4[0] = cos(((i+1)*2.0+1.0)*M_PI/DetailLevel);
    N4[1] = 0.0; P4[1] = 1.0;
    N4[2] = P4[2] = sin(((i+1)*2.0+1.0)*M_PI/DetailLevel);
    if (cylinder->Trans)
    {
      MEvaluatePoint(P1, cylinder->Trans, P1);
      MEvaluateVector(N1, cylinder->Trans, N1);
      MEvaluatePoint(P2, cylinder->Trans, P2);
      MEvaluateVector(N2, cylinder->Trans, N2);
      MEvaluatePoint(P3, cylinder->Trans, P3);
      MEvaluateVector(N3, cylinder->Trans, N3);
      MEvaluatePoint(P4, cylinder->Trans, P4);
      MEvaluateVector(N4, cylinder->Trans, N4);
    }
    AddTriangleToObjectMesh(cylinder, P1,P2,P3, N1,N2,N3);
    AddTriangleToObjectMesh(cylinder, P2,P3,P4, N2,N3,N4);
    if (!(ccylinder->Open))
    {
      AddTriangleToObjectMesh(cylinder, PO1,P1,P3, NO1, NO1, NO1);
      AddTriangleToObjectMesh(cylinder, PO2,P2,P4, NO2, NO2, NO2);
    }
  }
}

/*****************************************************************************
*  giram_cylinder_inside
******************************************************************************/
static gboolean giram_cylinder_inside(ObjectStruct *cylinder, double x, double y, double z)
{ /* XXX: clipped_by */
  Vector Vect;

  Vect[0] = x;
  Vect[1] = y;
  Vect[2] = z;
  if (cylinder->Trans)
  {
    MInverseEvaluatePoint(Vect, cylinder->Trans, Vect);
  }
  if (Vect[1] < 0.0 || Vect[1] > 1.0)
    return cylinder->Inverse;
  if (Vect[0]*Vect[0]+Vect[2]*Vect[2] > 1.0)
    return cylinder->Inverse;
  else
    return !(cylinder->Inverse);
}

/*****************************************************************************
*  giram_cylinder_is_intersection
******************************************************************************/
static gboolean giram_cylinder_is_intersection(ObjectStruct *cylinder, Vector origin, Vector direction)
{ /* XXX: clipped_by */
  Vector Org, Dir;
  int i;
  double A, B, C, D;
  double t1, t2, z1, z2;
  CylinderStruct *ccylinder = (CylinderStruct *)cylinder;

  if (cylinder->Trans)
  {
    MInverseEvaluatePoint(Org, cylinder->Trans, origin);
    MInverseEvaluateVector(Dir, cylinder->Trans, direction);
  } else
  {
    for (i=0 ; i<3 ; i++)
    {
      Org[i] = origin[i];
      Dir[i] = direction[i];
    }
  }
  if (Dir[0] == 0.0 && Dir[2] == 0.0)
  {
    if (ccylinder->Open)
      return FALSE;
    else
    {
      if (Org[0]*Org[0]+Org[2]*Org[2] < 1.0)
        return TRUE;
      else
        return FALSE;
    }
  }
  A = Dir[0]*Dir[0] + Dir[2]*Dir[2];
  B = 2.0*(Org[0]*Dir[0] + Org[2]*Dir[2]);
  C = Org[0]*Org[0] + Org[2]*Org[2] - 1.0;
  D = B*B - 4.0*A*C;
  if (D < 0.0)
    return FALSE;
  t1 = (-B-sqrt(D)) / 2.0 / A;
  z1 = Org[1] + t1*Dir[1];
  t2 = (-B+sqrt(D)) / 2.0 / A;
  z2 = Org[1] + t2*Dir[1];
  if ((z1>=0.0 && z1<=1.0) || (z2>=0.0 && z2<=1.0))
    return TRUE;
  if ((z1>1.0) && (z2>1.0))
    return FALSE;
  if ((z1<0.0) && (z2<0.0))
    return FALSE;
  return !(ccylinder->Open);
}

/*****************************************************************************
*  giram_cylinder_click_in_XY
******************************************************************************/
static gboolean giram_cylinder_click_in_XY(ObjectStruct *cylinder, double x, double y)
{
  Vector Origin = {0.0, 0.0, 0.0, 0.0, 0.0};
  Vector Direction = {0.0, 0.0, 1.0, 0.0, 0.0};

  if (cylinder->Trans)
  {
    Origin[0] = x;
    Origin[1] = y;
    return giram_object_is_intersection(cylinder, Origin, Direction);
  } else if ( (x>=-1.0) && (x<=1.0) && (y>=0.0) && (y<=1.0) )
    return TRUE;
  else
    return FALSE;
}

/****************************************************************************
*  giram_cylinder_click_in_XZ
*****************************************************************************/
static gboolean giram_cylinder_click_in_XZ(ObjectStruct *cylinder, double x, double z)
{
  Vector Origin = {0.0, 0.0, 0.0, 0.0, 0.0};
  Vector Direction = {0.0, 1.0, 0.0, 0.0, 0.0};

  if (cylinder->Trans)
  {
    Origin[0] = x;
    Origin[2] = z;
    return giram_object_is_intersection(cylinder, Origin, Direction);
  } else if ( x*x+z*z <= 1.0 )
    return TRUE;
  else
    return FALSE;
}

/****************************************************************************
*  giram_cylinder_click_in_ZY
*****************************************************************************/
static gboolean giram_cylinder_click_in_ZY(ObjectStruct *cylinder, double z, double y)
{
  Vector Origin = {0.0, 0.0, 0.0, 0.0, 0.0};
  Vector Direction = {1.0, 0.0, 0.0, 0.0, 0.0};

  if (cylinder->Trans)
  {
    Origin[2] = z;
    Origin[1] = y;
    return giram_object_is_intersection(cylinder, Origin, Direction);
  } else if ( (z>=-1.0) && (z<=1.0) && (y>=0.0) && (y<=1.0) )
    return TRUE;
  else
    return FALSE;
}

/*****************************************************************************
*  giram_cylinder_find_intersection_segment
******************************************************************************/
static gboolean giram_cylinder_find_intersection_segment(ObjectStruct *cylinder,
                                                         Vector in_point, Vector out_point,
                                                         Vector inter_point, Vector inter_norm)
{
  int i;
  Vector InP, OutP;
  double A, B, C, Delta, t;
  Vector Tmp;

  if (cylinder->Trans)
  {
    MInverseEvaluatePoint(InP, cylinder->Trans, in_point);
    MInverseEvaluatePoint(OutP, cylinder->Trans, out_point);
  } else
  {
    for (i=0 ; i<3 ; i++)
    {
      InP[i] = in_point[i];
      OutP[i] = out_point[i];
    }
  }
  A = (OutP[0]-InP[0])*(OutP[0]-InP[0]) + (OutP[2]-InP[2])*(OutP[2]-InP[2]);
  B = 2.0*(InP[0]*(OutP[0]-InP[0]) + InP[2]*(OutP[2]-InP[2]));
  C = InP[0]*InP[0] + InP[2]*InP[2] - 1;
  Delta = B*B-4*A*C;
  if (Delta < 0.0)
  { /* The Ray miss the cylinder or is parallel to the cylinder axis */
    if (fabs(A) < 10e-6)
    { /* The Ray is parallel to the Cylinder axis */
      if (C < 0.0)
      {
        if ( ( (InP[1] >= 0.0) && (InP[1] <= 1.0) && (OutP[1] >= 1.0) ) ||
             ( (OutP[1] >= 0.0) && (OutP[1] <= 1.0) && (InP[1] >= 1.0) ) )
        { /* The Ray Intersect the Top */
          inter_point[0] = InP[0]; inter_point[1] = 1.0; inter_point[2] = InP[2];
          inter_norm[0] = 0.0; inter_norm[1] = 1.0; inter_norm[2] = 0.0;
          if (cylinder->Trans)
          {
            MEvaluatePoint(inter_point, cylinder->Trans, inter_point);
            MEvaluateVector(inter_norm, cylinder->Trans, inter_norm);
          }
          return TRUE;
        }
        if ( ( (InP[1] >= 0.0) && (InP[1] <= 1.0) && (OutP[1] <= 0.0) ) ||
             ( (OutP[1] >= 0.0) && (OutP[1] <= 1.0) && (InP[1] <= 0.0) ) )
        { /* The Ray Intersect the Bottom */
          inter_point[0] = InP[0]; inter_point[1] = 0.0; inter_point[2] = InP[2];
          inter_norm[0] = 0.0; inter_norm[1] = -1.0; inter_norm[2] = 0.0;
          if (cylinder->Trans)
          {
            MEvaluatePoint(inter_point, cylinder->Trans, inter_point);
            MEvaluateVector(inter_norm, cylinder->Trans, inter_norm);
          }
          return TRUE;
        }
      }
    }
    return FALSE;
  }
  t = (-B - sqrt(Delta)) / (2.0 * A);
  if ((t >= 0.0) && (t <= 1.0))
  {
    Tmp[0] = InP[0] + (OutP[0] - InP[0]) * t;
    Tmp[1] = InP[1] + (OutP[1] - InP[1]) * t;
    Tmp[2] = InP[2] + (OutP[2] - InP[2]) * t;
  } else
  {
    t = (-B + sqrt(Delta)) / (2.0 * A);
    if ((t >= 0.0) && (t <= 1.0))
    {
      Tmp[0] = InP[0] + (OutP[0] - InP[0]) * t;
      Tmp[1] = InP[1] + (OutP[1] - InP[1]) * t;
      Tmp[2] = InP[2] + (OutP[2] - InP[2]) * t;
    } else
      return FALSE;
  }
  if ( (Tmp[1] >= 0.0) && (Tmp[1] <= 1.0) )
  {
    inter_point[0] = Tmp[0]; inter_point[1] = Tmp[1]; inter_point[2] = Tmp[2];
    inter_norm[0] = Tmp[0]; inter_norm[1] = 0.0; inter_norm[2] = Tmp[2];
    if (cylinder->Trans)
    {
      MEvaluatePoint(inter_point, cylinder->Trans, inter_point);
      MEvaluateVector(inter_norm, cylinder->Trans, inter_norm);
    }
    return TRUE;
  }
  return FALSE;
}

/****************************************************************************/
double Xorg, Yorg, Xold, Yold;
GdkColormap *colormap;
GdkGC *gcPurple;
GdkColor Purple;
guint id;

#define NONE   1
#define MOVING 2
int CYLINDER_STATE = NONE;

/*****************************************************************************
*  ToolInitNewCylinder
******************************************************************************/
void ToolInitNewCylinder(GtkWidget *DrawingArea, GdkEventButton *bevent)
{
  ViewUserDataStruct *UserData;
  GdkColormap *colormap;
  GtkWidget *StatusBar;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  id = gtk_statusbar_get_context_id(GTK_STATUSBAR(StatusBar), "new cylinder");
  MouseToReal(bevent->x, bevent->y, DrawingArea, &Xorg, &Yorg);
  colormap = gdk_window_get_colormap(DrawingArea->window);
  gcPurple = gdk_gc_new(DrawingArea->window);
  Purple.red = 65535; Purple.green = 0; Purple.blue = 65535;
  /* XXX: the color should be settable... */
  gdk_color_alloc(colormap, &Purple);
  gdk_gc_set_foreground(gcPurple, &Purple);
  gdk_gc_set_function(gcPurple, GDK_XOR);

  Xold = Xorg; Yold = Yorg;

  CYLINDER_STATE = MOVING;
  gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, _("Cylinder: "));
  gdk_pointer_grab(DrawingArea->window, FALSE,
                   GDK_POINTER_MOTION_MASK | GDK_BUTTON_RELEASE_MASK,
                   NULL, NULL, bevent->time);
}

/*****************************************************************************
*  ToolMotionNewCylinder
******************************************************************************/
void ToolMotionNewCylinder(GtkWidget *DrawingArea, GdkEvent *event)
{
  GdkEventMotion *Mev = (GdkEventMotion *)event;
  ViewUserDataStruct *UserData;
  GtkWidget *StatusBar;
  double Zoom, Xoff, Yoff;
  gchar *Message;
  double x1, y1, x2, y2;
  double gx1, gy1;
  double Length, Radius;
  Vector From;

  if (CYLINDER_STATE != MOVING)
    return;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
  gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);

  if (Xorg < Xold)
  {
    x1 = Xorg-(Xold-Xorg); x2 = Xold;
  } else
  {
    x1 = Xold; x2 = Xorg+Xorg-Xold;
  }
  y1 = MIN(Yorg, Yold); y2 = MAX(Yorg, Yold);
  gdk_draw_rectangle(DrawingArea->window, gcPurple, FALSE,
                     DrawingArea->allocation.width/2+x1*Zoom, 
                     DrawingArea->allocation.height/2-y2*Zoom, 
                     (x2-x1)*Zoom, (y2-y1)*Zoom);
  MouseToReal(Mev->x, Mev->y, DrawingArea, &Xold, &Yold);
  if (Xorg < Xold)
  {
    x1 = Xorg-(Xold-Xorg); x2 = Xold;
  } else
  {
    x1 = Xold; x2 = Xorg+Xorg-Xold;
  }
  y1 = MIN(Yorg, Yold); y2 = MAX(Yorg, Yold);
  gdk_draw_rectangle(DrawingArea->window, gcPurple, FALSE,
                     DrawingArea->allocation.width/2+x1*Zoom, 
                     DrawingArea->allocation.height/2-y2*Zoom, 
                     (x2-x1)*Zoom, (y2-y1)*Zoom);
                                
  gx1 = Xoff+(Xorg-DrawingArea->allocation.width /2)/Zoom;
  Radius = fabs(Xorg-Mev->x)/Zoom;
      
  gy1 = MIN(Yorg, Yold);
  Length = fabs(Yorg-Yold);
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      From[0] = Xorg;
      From[1] = gy1;
      From[2] = 0.0;
      break;
    case ORTHO_XZ_CAMERA:
      From[0] = Xorg;
      From[1] = 0.0;
      From[2] = gy1;
      break;
    case ORTHO_ZY_CAMERA:
      From[0] = 0.0;
      From[1] = gy1;
      From[2] = Xorg;
      break;
  }
  Message = g_strdup_printf(_("Cylinder: From <%g, %g, %g>, Length: %g, Radius: %g"),
                            From[0], From[1], From[2], Length, Radius);
  gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, Message);
  g_free(Message);
}

/*****************************************************************************
*  ToolReleaseNewCylinder
******************************************************************************/
void ToolReleaseNewCylinder(GtkWidget *DrawingArea, GdkEvent *event)
{
  GdkEventButton *Bev = (GdkEventButton *)event;
  ViewUserDataStruct *UserData;
  GtkWidget          *StatusBar;
  double              Zoom, Xoff, Yoff;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;
  ObjectStruct       *cylinder;
  CylinderStruct     *ccylinder;
  double              gx1, gy1, gy2;
  double              Radius;
  Vector              Base, Apex;

  if (CYLINDER_STATE != MOVING)
    return;
  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;

  gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);
  
  MouseToReal(Bev->x, Bev->y, DrawingArea, &Xold, &Yold);
  
  gx1 = Xorg;
  Radius = fabs(Xorg-Xold);
      
  gy1 = MIN(Yorg, Yold);
  gy2 = MAX(Yorg, Yold);

  if ((Radius > 10e-6) && (gy2-gy1 > 10e-6))
  {
    switch (UserData->CameraStyle)
    {
      case ORTHO_XY_CAMERA:
        Base[0] = gx1; Base[1] = gy1; Base[2] = 0.0;
        Apex[0] = gx1; Apex[1] = gy2; Apex[2] = 0.0;
        break;
      case ORTHO_XZ_CAMERA:
        Base[0] = gx1; Base[1] = 0.0; Base[2] = gy1;
        Apex[0] = gx1; Apex[1] = 0.0; Apex[2] = gy2;
        break;
      case ORTHO_ZY_CAMERA:
        Base[0] = 0.0; Base[1] = gy1; Base[2] = gx1;
        Apex[0] = 0.0; Apex[1] = gy2; Apex[2] = gx1;
        break;
    }
    cylinder = giram_cylinder_new(Base, Apex, Radius);
    ccylinder = (CylinderStruct *)cylinder;
    ccylinder->Open = CylinderOpenFlag;
    UserData->FrameView->all_objects = g_slist_append(UserData->FrameView->all_objects,
                                                      cylinder);
    giram_object_build_triangle_mesh(cylinder);
  }
  CYLINDER_STATE = NONE;

  gdk_gc_destroy(gcPurple);
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, UserData->FrameView);
  gdk_pointer_ungrab(Bev->time);
}

/*****************************************************************************
*  ToggleCylinderOpen
******************************************************************************/
void ToggleCylinderOpen(GtkWidget *Check)
{
  CylinderOpenFlag = !CylinderOpenFlag;
}

/*****************************************************************************
*  BuildCylinderOptions
******************************************************************************/
GtkWidget *BuildCylinderOptions(GtkWidget *VBox)
{
  GtkWidget *vbox;
  GtkWidget *CheckOpen;
  GtkWidget *label;

  /*  the main vbox  */
  vbox = gtk_vbox_new (FALSE, 1);
  /*  the main label  */
  label = gtk_label_new(_("Cylinder"));
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

  CheckOpen = gtk_check_button_new_with_label(_("Open-ended"));
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(CheckOpen), CylinderOpenFlag);
  gtk_signal_connect_object(GTK_OBJECT(CheckOpen), "toggled",
                            GTK_SIGNAL_FUNC(ToggleCylinderOpen), NULL);

  gtk_box_pack_start(GTK_BOX(vbox), CheckOpen, FALSE, FALSE, 0);

  gtk_box_pack_start(GTK_BOX(VBox), vbox, TRUE, TRUE, 0);
  gtk_widget_show_all(vbox);
  return vbox;
}

/*****************************************************************************
*  giram_cylinder_new
******************************************************************************/
ObjectStruct *giram_cylinder_new(Vector Base, Vector Apex, double Radius)
{
  ObjectStruct *cylinder;
  CylinderStruct *ccylinder;
  static GiramObjectClass *cylinder_klass = NULL;
  Vector LocalVector1, LocalVector2, Translate, Scale, N, U1, U2;
  double Length, ULength;
  TransformStruct Transform;

  if (cylinder_klass == NULL)
  {
    cylinder_klass = giram_object_class_new();

    cylinder_klass->build_triangle_mesh       = giram_cylinder_build_triangle_mesh;
    cylinder_klass->inside                    = giram_cylinder_inside;
    cylinder_klass->is_intersection           = giram_cylinder_is_intersection;
    cylinder_klass->click_in_XY               = giram_cylinder_click_in_XY;
    cylinder_klass->click_in_XZ               = giram_cylinder_click_in_XZ;
    cylinder_klass->click_in_ZY               = giram_cylinder_click_in_ZY;
    cylinder_klass->find_intersection_segment = giram_cylinder_find_intersection_segment;
  }
  ccylinder = g_new(CylinderStruct, 1);
  cylinder = (ObjectStruct *)ccylinder;
  InitObject(cylinder);
  cylinder->Type = CYLINDER_OBJECT;
  cylinder->klass = cylinder_klass;
  LocalVector1[0] = Base[0];
  LocalVector1[1] = Base[1];
  LocalVector1[2] = Base[2];
  LocalVector2[0] = Apex[0];
  LocalVector2[1] = Apex[1];
  LocalVector2[2] = Apex[2];
  Translate[0] = LocalVector1[0];
  Translate[1] = LocalVector1[1];
  Translate[2] = LocalVector1[2];
  LocalVector2[0] -= Translate[0];
  LocalVector2[1] -= Translate[1];
  LocalVector2[2] -= Translate[2];

  Length = VLength(LocalVector2);
  Scale[0] = Radius;
  Scale[1] = Length;
  Scale[2] = Radius;
  giram_object_scale(cylinder, Scale);

  N[0] = LocalVector2[0] / Length;
  N[1] = LocalVector2[1] / Length;
  N[2] = LocalVector2[2] / Length;
  if ( (fabs(N[0])>fabs(N[1])) && (fabs(N[0])>fabs(N[2])) )
  { /* The cylinder axis is mainly upon the X axis */
    U1[0] = -N[1]/N[0];
    U1[1] = 1;
    U1[2] = 0;
  } else if ( (fabs(N[1])>fabs(N[0])) && (fabs(N[1])>fabs(N[2])) )
  { /* The cylinder axis is mainly upon the Y axis */
    U1[0] = 1;
    U1[1] = -N[0]/N[1];
    U1[2] = 0;
  } else
  { /* The cylinder axis is mainly upon the Z axis */
    U1[0] = 0;
    U1[1] = 1;
    U1[2] = -N[1]/N[2];
  }
  ULength = VLength(U1);
  U1[0]/=ULength; U1[1]/=ULength; U1[2]/=ULength;
  VCross(U2, N, U1);
  ULength = VLength(U2);
  U2[0]/=ULength; U2[1]/=ULength; U2[2]/=ULength;
  Transform.Direct[0][0] = U1[0];
  Transform.Direct[0][1] = U1[1];
  Transform.Direct[0][2] = U1[2];
  Transform.Direct[0][3] = 0.0;
  Transform.Direct[1][0] = N[0];
  Transform.Direct[1][1] = N[1];
  Transform.Direct[1][2] = N[2];
  Transform.Direct[1][3] = 0.0;
  Transform.Direct[2][0] = U2[0];
  Transform.Direct[2][1] = U2[1];
  Transform.Direct[2][2] = U2[2];
  Transform.Direct[2][3] = 0.0;
  Transform.Direct[3][0] = 0.0;
  Transform.Direct[3][1] = 0.0;
  Transform.Direct[3][2] = 0.0;
  Transform.Direct[3][3] = 1.0;
  MInverse(Transform.Inverse, Transform.Direct);
  ComposeTrans(cylinder->Trans, &Transform);
  giram_object_translate(cylinder, Translate);
  ccylinder->Open = CylinderOpenFlag;
  return cylinder;
}

/*****************************************************************************
*  PDBCreateCylinder
******************************************************************************/
int PDBCreateCylinder(int *Id, Vector *Base, Vector *Apex, double *Radius)
{
  ObjectStruct    *Cylinder;
  FrameStruct     *LocalFrame;
  GSList          *tmp_list;
  FrameViewStruct *TmpView;
    
  Cylinder = giram_cylinder_new(*Base, *Apex, *Radius);
  for (LocalFrame = FirstFrame ;
       LocalFrame->Id != *Id ;
       LocalFrame = LocalFrame->Next)
    ;
  LocalFrame->all_objects = g_slist_append(LocalFrame->all_objects, Cylinder);
  giram_object_build_triangle_mesh(Cylinder);
  for (tmp_list = LocalFrame->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, LocalFrame);
  return Cylinder->Id;
}

