/* disc.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "giram.h"
#include "trimesh.h"
#include "csgtree.h"
#include "object.h"
#include "snap.h"
#include "utils.h"
#include "disc.h"

#include "giramintl.h"

int DiscWithHoleFlag;

static void giram_disc_build_triangle_mesh(ObjectStruct *disc);
static gboolean giram_disc_inside(ObjectStruct *disc, double x, double y, double z);
static gboolean giram_disc_is_intersection(ObjectStruct *disc, Vector org, Vector dir);
static gboolean giram_disc_find_intersection_segment(ObjectStruct *disc,
                                                    Vector in_point, Vector out_point,
                                                    Vector inter_point, Vector inter_norm);

/*****************************************************************************
*  giram_disc_build_triangle_mesh
******************************************************************************/
static void giram_disc_build_triangle_mesh(ObjectStruct *disc)
{ /* XXX: clipped_by */
  DiscStruct *ddisc = (DiscStruct *)disc;
  Vector P1, P2, P3, P4;
  Vector N;
  int i;

  disc->DetailLevel = 16;

  if (disc->FirstTriangle)
    DestroyObjectTriangleMesh(disc);
  N[0] = 0.0; N[1] = 1.0; N[2] = 0.0;
  if (disc->Trans)
  {
    MEvaluateVector(N, disc->Trans, N);
  }
  for (i=0 ; i<disc->DetailLevel ; i++)
  {
    P1[0] = ddisc->HoleRadius*cos(i*2.0*M_PI/disc->DetailLevel);
    P1[1] = 0.0;
    P1[2] = ddisc->HoleRadius*sin(i*2.0*M_PI/disc->DetailLevel);
    P2[0] = ddisc->Radius*cos((i*2.0+1.0)*M_PI/disc->DetailLevel);
    P2[1] = 0.0;
    P2[2] = ddisc->Radius*sin((i*2.0+1.0)*M_PI/disc->DetailLevel);
    P3[0] = ddisc->HoleRadius*cos((i+1)*2.0*M_PI/disc->DetailLevel);
    P3[1] = 0.0;
    P3[2] = ddisc->HoleRadius*sin((i+1)*2.0*M_PI/disc->DetailLevel);
    P4[0] = ddisc->Radius*cos(((i+1)*2.0+1.0)*M_PI/disc->DetailLevel);
    P4[1] = 0.0;
    P4[2] = ddisc->Radius*sin(((i+1)*2.0+1.0)*M_PI/disc->DetailLevel);
    if (disc->Trans)
    {
      MEvaluatePoint(P1, disc->Trans, P1);
      MEvaluatePoint(P2, disc->Trans, P2);
      MEvaluatePoint(P3, disc->Trans, P3);
      MEvaluatePoint(P4, disc->Trans, P4);
    }
    if (ddisc->HoleRadius > 10e-6)
      AddTriangleToObjectMesh(disc, P1,P2,P3, N,N,N);
    AddTriangleToObjectMesh(disc, P2,P3,P4, N,N,N);
  }
}

/*****************************************************************************
*  giram_disc_inside
******************************************************************************/
static gboolean giram_disc_inside(ObjectStruct *disc, double x, double y, double z)
{
  return FALSE; /* We cannot be "Inside" a flat disc... */
}

/*****************************************************************************
*  giram_disc_find_intersection_segment
******************************************************************************/
static gboolean giram_disc_find_intersection_segment(ObjectStruct *disc,
                                                    Vector in_point, Vector out_point,
                                                    Vector inter_point, Vector inter_norm)
{
  return FALSE; /* This fonction cannot be called anyway... */
}

/*****************************************************************************
*  giram_disc_is_intersection FIXME
******************************************************************************/
static gboolean giram_disc_is_intersection(ObjectStruct *disc, Vector origin, Vector direction)
{
  Vector Org, Dir;
  int i;

  if (disc->Trans)
  {
    MInverseEvaluatePoint(Org, disc->Trans, origin);
    MInverseEvaluateVector(Dir, disc->Trans, direction);
  } else
  {
    for (i=0 ; i<3 ; i++)
    {
      Org[i] = origin[i];
      Dir[i] = direction[i];
    }
  }
  return FALSE;
}

/****************************************************************************/
double Xorg, Yorg, Radius, HoleRadius;
GdkColormap *colormap;
GdkGC *gcPurple;
GdkColor Purple;
guint id;
Vector Center;

#define NONE           1
#define MOVING         2
#define BETWEEN_RADIUS 3
#define HOLE_RADIUS    4

int DISC_STATE = NONE;

/*****************************************************************************
*  ToolInitNewDisc
******************************************************************************/
void ToolInitNewDisc(GtkWidget *DrawingArea, GdkEventButton *bevent)
{
  ViewUserDataStruct *UserData;
  GdkColormap *colormap;
  GtkWidget *StatusBar;
  double Zoom, Xoff, Yoff;

  if (DISC_STATE == BETWEEN_RADIUS)
  {
    DISC_STATE = HOLE_RADIUS;
    return;
  }
  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
  id = gtk_statusbar_get_context_id(GTK_STATUSBAR(StatusBar), "new disc");
  MouseToReal(bevent->x, bevent->y, DrawingArea, &Xorg, &Yorg);

  colormap = gdk_window_get_colormap(DrawingArea->window);
  gcPurple = gdk_gc_new(DrawingArea->window);
  Purple.red = 65535; Purple.green = 0; Purple.blue = 65535;
  /* XXX: the color should be settable... */
  gdk_color_alloc(colormap, &Purple);
  gdk_gc_set_foreground(gcPurple, &Purple);
  gdk_gc_set_function(gcPurple, GDK_XOR);

  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      Center[0] = Xorg;
      Center[1] = Yorg;
      Center[2] = 0.0;
      break;
    case ORTHO_XZ_CAMERA:
      Center[0] = Xorg;
      Center[1] = 0.0;
      Center[2] = Yorg;
      break;
    case ORTHO_ZY_CAMERA:
      Center[0] = 0.0;
      Center[1] = Yorg;
      Center[2] = Xorg;
      break;
  }
  Radius = 0.0;

  DISC_STATE = MOVING;
  gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, _("Disc: "));
  gdk_pointer_grab(DrawingArea->window, FALSE,
                   GDK_BUTTON_PRESS_MASK |
                   GDK_POINTER_MOTION_MASK |
                   GDK_BUTTON_RELEASE_MASK,
                   NULL, NULL, bevent->time);
}

/*****************************************************************************
*  ToolMotionNewDisc
******************************************************************************/
void ToolMotionNewDisc(GtkWidget *DrawingArea, GdkEvent *event)
{
  GdkEventMotion *Mev = (GdkEventMotion *)event;
  ViewUserDataStruct *UserData;
  GtkWidget *StatusBar;
  double Zoom, Xoff, Yoff, Xold, Yold;
  char *Message;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
  if (DISC_STATE == MOVING)
  {
    gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);

    gdk_draw_arc(DrawingArea->window, gcPurple, FALSE,
                 DrawingArea->allocation.width/2+(Xorg-Radius-Xoff)*Zoom,
                 DrawingArea->allocation.height/2-(Yorg+Radius-Yoff)*Zoom,
                 Radius*2.0*Zoom, Radius*2.0*Zoom,  0, 64*360);
    MouseToReal(Mev->x, Mev->y, DrawingArea, &Xold, &Yold);
    Radius = sqrt((Xold-Xorg)*(Xold-Xorg)+(Yold-Yorg)*(Yold-Yorg));
    gdk_draw_arc(DrawingArea->window, gcPurple, FALSE,
                 DrawingArea->allocation.width/2+(Xorg-Radius-Xoff)*Zoom,
                 DrawingArea->allocation.height/2-(Yorg+Radius-Yoff)*Zoom,
                 Radius*2.0*Zoom, Radius*2.0*Zoom,  0, 64*360);
    Message = g_strdup_printf(_("Disc: Center <%g, %g, %g>, Radius: %g"),
                              Center[0], Center[1], Center[2], Radius);
    gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, Message);
    g_free(Message);
  } else if ((DISC_STATE == HOLE_RADIUS) || (DISC_STATE == BETWEEN_RADIUS))
  {
    gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);

    gdk_draw_arc(DrawingArea->window, gcPurple, FALSE,
                 DrawingArea->allocation.width/2+(Xorg-HoleRadius-Xoff)*Zoom,
                 DrawingArea->allocation.height/2-(Yorg+HoleRadius-Yoff)*Zoom,
                 HoleRadius*2.0*Zoom, HoleRadius*2.0*Zoom, 0,64*360);
    MouseToReal(Mev->x, Mev->y, DrawingArea, &Xold, &Yold);
    HoleRadius = sqrt((Xold-Xorg)*(Xold-Xorg)+(Yold-Yorg)*(Yold-Yorg));
    gdk_draw_arc(DrawingArea->window, gcPurple, FALSE,
                 DrawingArea->allocation.width/2+(Xorg-HoleRadius-Xoff)*Zoom,
                 DrawingArea->allocation.height/2-(Yorg+HoleRadius-Yoff)*Zoom,
                 HoleRadius*2.0*Zoom, HoleRadius*2.0*Zoom, 0,64*360);
    Message = g_strdup_printf(_("Disc: Center <%g, %g, %g>, Radius: %g, Hole Radius: %g"),
            Center[0], Center[1], Center[2], 
            MAX(Radius, HoleRadius),
            MIN(Radius, HoleRadius));
    gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, Message);
    g_free(Message);
  }
}

/*****************************************************************************
*  ToolReleaseNewDisc
******************************************************************************/
void ToolReleaseNewDisc(GtkWidget *DrawingArea, GdkEvent *event)
{
  GdkEventButton *Bev = (GdkEventButton *)event;
  ViewUserDataStruct *UserData;
  GtkWidget          *StatusBar;
  double              Zoom, Xoff, Yoff;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;
  ObjectStruct       *Disc;
  Vector              Normal;

  if (DISC_STATE == MOVING)
  {
    DISC_STATE = BETWEEN_RADIUS;
    HoleRadius = 0.0;
    return;
  }
  if (DISC_STATE != HOLE_RADIUS)
    return;
  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
  gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      Normal[0] = 0.0;
      Normal[1] = 0.0;
      Normal[2] = 1.0;
      break;
    case ORTHO_XZ_CAMERA:
      Normal[0] = 0.0;
      Normal[1] = 1.0;
      Normal[2] = 0.0;
      break;
    case ORTHO_ZY_CAMERA:
      Normal[0] = 1.0;
      Normal[1] = 0.0;
      Normal[2] = 0.0;
      break;
  }
  Disc = giram_disc_new(Center, Normal, Radius, HoleRadius);
  UserData->FrameView->all_objects = g_slist_append(UserData->FrameView->all_objects,
                                                    Disc);
  giram_object_build_triangle_mesh(Disc);
  DISC_STATE = NONE;
  gdk_gc_destroy(gcPurple);
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, UserData->FrameView);
  gdk_pointer_ungrab(Bev->time);
}

/*****************************************************************************
*  ToggleDiscWithHole
******************************************************************************/
void ToggleDiscWithHole(GtkWidget *Check)
{
  DiscWithHoleFlag = !DiscWithHoleFlag;
}

/*****************************************************************************
*  BuildDiscOptions
******************************************************************************/
GtkWidget *BuildDiscOptions(GtkWidget *VBox)
{
  GtkWidget *vbox;
  GtkWidget *Check;
  GtkWidget *label;

  /*  the main vbox  */
  vbox = gtk_vbox_new (FALSE, 1);
  /*  the main label  */
  label = gtk_label_new(_("Disc"));
  gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

  Check = gtk_check_button_new_with_label(_("Disc with Hole"));
  DiscWithHoleFlag = TRUE;
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(Check), DiscWithHoleFlag);
  gtk_signal_connect_object(GTK_OBJECT(Check), "toggled",
                            GTK_SIGNAL_FUNC(ToggleDiscWithHole), NULL);
  gtk_box_pack_start(GTK_BOX(vbox), Check, FALSE, FALSE, 0);
                     gtk_widget_show(Check);

  gtk_box_pack_start(GTK_BOX(VBox), vbox, TRUE, TRUE, 0);
                     gtk_widget_show(vbox);
  return vbox;
}

/*****************************************************************************
*  giram_disc_new
******************************************************************************/
ObjectStruct *giram_disc_new(Vector center, Vector Norm, double Radius, double HoleRadius)
{
  ObjectStruct *disc;
  DiscStruct *ddisc;
  static GiramObjectClass *disc_klass = NULL;
  Vector Normal, U1, U2;
  double Length, ULength;

  if (disc_klass == NULL)
  {
    disc_klass = giram_object_class_new();

    disc_klass->build_triangle_mesh       = giram_disc_build_triangle_mesh;
    disc_klass->inside                    = giram_disc_inside;
    disc_klass->is_intersection           = giram_disc_is_intersection;
    disc_klass->find_intersection_segment = giram_disc_find_intersection_segment;
  }
  ddisc = g_new(DiscStruct, 1);
  disc = (ObjectStruct *)ddisc;
  InitObject(disc);
  disc->Type = DISC_OBJECT;
  disc->klass = disc_klass;
  ddisc->Radius = MAX(Radius, HoleRadius);
  ddisc->HoleRadius = MIN(Radius, HoleRadius);

  disc->Trans = g_malloc(sizeof(TransformStruct));
  Normal[0] = Norm[0]; Normal[1] = Norm[1]; Normal[2] = Norm[2];
  Length = VLength(Normal);
  Normal[0] /= Length;
  Normal[1] /= Length;
  Normal[2] /= Length;
  if ( (fabs(Normal[0])>fabs(Normal[1])) && (fabs(Normal[0])>fabs(Normal[2])) )
  { /* The normal is mainly upon the X axis */
    U1[0] = -Normal[1]/Normal[0];
    U1[1] = 1;
    U1[2] = 0;
  } else if ( (fabs(Normal[1])>fabs(Normal[0])) && (fabs(Normal[1])>fabs(Normal[2])) )
  { /* The normal is mainly upon the Y axis */
    U1[0] = 1;
    U1[1] = -Normal[0]/Normal[1];
    U1[2] = 0;
  } else
  { /* The normal axis is mainly upon the Z axis */
    U1[0] = 0;
    U1[1] = 1;
    U1[2] = -Normal[1]/Normal[2];
  }
  ULength = VLength(U1);
  U1[0]/=ULength; U1[1]/=ULength; U1[2]/=ULength;
  VCross(U2, Normal, U1);
  ULength = VLength(U2);
  U2[0]/=ULength; U2[1]/=ULength; U2[2]/=ULength;

  disc->Trans->Direct[0][0] = U1[0];
  disc->Trans->Direct[0][1] = U1[1];
  disc->Trans->Direct[0][2] = U1[2];
  disc->Trans->Direct[0][3] = 0.0;
  disc->Trans->Direct[1][0] = Normal[0];
  disc->Trans->Direct[1][1] = Normal[1];
  disc->Trans->Direct[1][2] = Normal[2];
  disc->Trans->Direct[1][3] = 0.0;
  disc->Trans->Direct[2][0] = U2[0];
  disc->Trans->Direct[2][1] = U2[1];
  disc->Trans->Direct[2][2] = U2[2];
  disc->Trans->Direct[2][3] = 0.0;
  disc->Trans->Direct[3][0] = 0.0;
  disc->Trans->Direct[3][1] = 0.0;
  disc->Trans->Direct[3][2] = 0.0;
  disc->Trans->Direct[3][3] = 1.0;
  MInverse(disc->Trans->Inverse, disc->Trans->Direct);
  giram_object_translate(disc, center);

  return disc;
}

/*****************************************************************************
*  PDBCreateDisc
******************************************************************************/
int PDBCreateDisc(int *Id, Vector *Center, Vector *pNormal,
                  double *Radius, double *HoleRadius)
{
  ObjectStruct    *disc;
  FrameStruct     *LocalFrame;
  GSList          *tmp_list;
  FrameViewStruct *TmpView;

  disc = giram_disc_new(*Center, *pNormal, *Radius, *HoleRadius);

  for (LocalFrame = FirstFrame ;
       LocalFrame->Id != *Id ;
       LocalFrame = LocalFrame->Next)
                  ;
  LocalFrame->all_objects = g_slist_append(LocalFrame->all_objects, disc);
  giram_object_build_triangle_mesh(disc);
  for (tmp_list = LocalFrame->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, LocalFrame);
  return disc->Id;
}

