/* dxf_write.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 * Copyright (C) 1999-2000 Noah Davis of VIEWS Net, Inc. <noah@viewsnet.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * This file is from Noah Davis.
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "giram.h"
#include "object.h"
#include "csgtree.h"
#include "view.h"
#include "camera.h"
#include "view_camera.h"
#include "frame.h"
#include "cylinder.h"

#include "dxf.h"

#include "giramintl.h"

typedef struct DXFPrivate
{
  FrameStruct *Frame;
  char *FileName;
  int SelectedOnly;
  GtkWidget *Text;
  int IncludeComment;
} DXFPrivate; 

/*****************************************************************************
*  dxf_save_file
*  Dumps out Giram triangle data in an ASCII (for now) DXF file
******************************************************************************/
int dxf_save_file(FrameStruct *frame, char *fname, int binary, int SelectedOnly,
                  char *Comment)
{
  FILE *fp;
  GSList *tmp_list;
  ObjectStruct *iterObject;
  ObjectSelection *iterSel;

  if (binary)
  {
    fprintf (stderr, "DXF_SAVE: Binary writes not implemented yet.\n");
    return 1;
  }

  if (SelectedOnly)
  {
    if (frame->Selection == NULL)
    {
      fprintf(stderr, "DXF_SAVE: Nothing to write!\n");
      return 1;
    }
  } else
  {
    if (frame->all_objects == NULL)
    {
      fprintf(stderr, "DXF_SAVE: Nothing to write!\n");
      return 1;
    }
  }
  fp = fopen(fname,"w");

  if (fp == NULL)
  {
    fprintf(stderr, "DXF_SAVE: Couldn't open %s for writing.\n", fname);
    return 1;
  }

  /* Okay, we have an open file, and at least one thing to write to it.
   * Lets start dumping...
   */


  /* First, the header, eschewing all but the ENTITIES section */
  if (!dxf_write_header(fp, binary, Comment))
  {
    fclose(fp);
    return 1;
  }

  if (SelectedOnly)
  {
    for (iterSel = frame->Selection ;
         iterSel->Next != frame->Selection ;
         iterSel = iterSel->Next)
    {
      if (!dxf_write_object(fp, iterSel->Object))
      {
        fclose(fp);
        return 1;
      }
    }
    if (!dxf_write_object(fp, iterSel->Object))
    {
      fclose(fp);
      return 1;
    }
  } else
  {
    for (tmp_list = frame->all_objects ; 
         tmp_list ; 
         tmp_list = g_slist_next(tmp_list))
    {
      iterObject = (ObjectStruct*)(tmp_list->data);
      if (!dxf_write_object(fp, iterObject))
      {
        fclose(fp);
        return 1;
      }
    }
  }

  if (!dxf_write_end(fp, binary))
  {
    fclose(fp);
    return 1;
  }

  fclose (fp);
  return 0;
}

/*****************************************************************************
*  dxf_write_header
******************************************************************************/
int dxf_write_header(FILE *fp, int bin, char *Comment)
{
  int strpos = 0;
  char CommentLine[256];
#if 0
  if (bin) ; /* Nothing for now. */
#endif
  /* The blank HEADER section */
  fprintf(fp, "0\nSECTION\n2\nHEADER\n0\nENDSEC\n");

  /* Comments out, 256 Bytes per line */
  if (Comment)
  {
    while (Comment[strpos])
    {
      int i;
      for (i=0; 
           (Comment[strpos]) && (Comment[strpos]!='\n') && (i < 255) ;
           i++, strpos++)
      {
        CommentLine[i] = Comment[strpos];
      }
      if (i<255) strpos++;
      CommentLine[i] = 0;
      fprintf(fp, "999\n%s\n", CommentLine);
    }
  }

  /* The blank HEADER section */
  fprintf(fp, "0\nSECTION\n2\nHEADER\n0\nENDSEC\n");

  /* Start the ENTITIES section */
  fprintf(fp, "0\nSECTION\n2\nENTITIES\n");

  return TRUE;
}

/*****************************************************************************
*  dxf_write_object
******************************************************************************/
int dxf_write_object(FILE *fp, ObjectStruct *obj)
{
  TriangleListStruct *tri;

  if ((tri=obj->FirstTriangle) == NULL)
    return TRUE;   /* Nothing to write */

  for ( ; tri ; tri=tri->Next)
  {
    /* The entity.. always a 3dface for now */
    fprintf(fp, "0\n3DFACE\n");

    /* The layer, blank until Giram supports it */
    fprintf(fp, "8\nGiramLayer\n");

    fprintf(fp, "10\n%f\n", tri->P1[0]);
    fprintf(fp, "20\n%f\n", tri->P1[1]);
    fprintf(fp, "30\n%f\n", tri->P1[2]);

    fprintf(fp, "11\n%f\n", tri->P2[0]);
    fprintf(fp, "21\n%f\n", tri->P2[1]);
    fprintf(fp, "31\n%f\n", tri->P2[2]);

    fprintf(fp, "12\n%f\n", tri->P3[0]);
    fprintf(fp, "22\n%f\n", tri->P3[1]);
    fprintf(fp, "32\n%f\n", tri->P3[2]);

    fprintf(fp, "13\n%f\n", tri->P3[0]);
    fprintf(fp, "23\n%f\n", tri->P3[1]);
    fprintf(fp, "33\n%f\n", tri->P3[2]);
  }
  return TRUE;
}

/*****************************************************************************
*  dxf_write_end
******************************************************************************/
int dxf_write_end(FILE *fp, int bin)
{
#if 0 
  if (bin) ; /* GNDN */
#endif
  fprintf(fp, "0\nENDSEC\n0\nEOF\n");

  return TRUE;
}

/*****************************************************************************
*  SaveDXFOK
******************************************************************************/
void SaveDXFOK(GtkObject *Dialog)
{
  DXFPrivate *DXFData;

  DXFData = gtk_object_get_data(Dialog, "DXFData");

  if (DXFData->IncludeComment)
    dxf_save_file(DXFData->Frame, DXFData->FileName, 0, DXFData->SelectedOnly,
                  gtk_editable_get_chars(GTK_EDITABLE(DXFData->Text),0,-1) );
  else
    dxf_save_file(DXFData->Frame, DXFData->FileName, 0, DXFData->SelectedOnly,
                  NULL);
  g_free(DXFData);
}

/*****************************************************************************
*  ToggleDXFIncludeComment
******************************************************************************/
void ToggleDXFIncludeComment(GtkObject *Dialog)
{
  DXFPrivate *DXFData;

  DXFData = gtk_object_get_data(Dialog, "DXFData");
  DXFData->IncludeComment = !DXFData->IncludeComment;
}

/*****************************************************************************
*  SaveDXFDialog
******************************************************************************/
void SaveDXFDialog(FrameStruct *Frame, char *FileName, int SelectedOnly)
{
  GtkWidget *Dialog;
  GtkWidget *Button;
  GtkWidget *CheckButton;
  GtkWidget *frame;
  GtkWidget *Table;
  GtkWidget *Text;
  GtkWidget *Vscrollbar;
  char *Comment = "This file was Generated by Giram-" VERSION "\n"
                  "If you experience any error, you can reach the\n"
                  "author of the Giram's DXF exporter at <adoyle@rimmer.viewsnet.com>\n";
  DXFPrivate *DXFData;

  DXFData = g_malloc(sizeof(DXFPrivate));
  DXFData->Frame = Frame;
  DXFData->FileName = g_strdup(FileName);
  
  DXFData->SelectedOnly = SelectedOnly;
  /* The Dialog Box for the Optionnal Comment */
  Dialog = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(Dialog), "Save as Jpeg");
  gtk_window_position(GTK_WINDOW(Dialog), GTK_WIN_POS_MOUSE);
  gtk_object_set_data(GTK_OBJECT(Dialog), "DXFData", DXFData);
  
  /* The "OK" Button */
  Button = gtk_button_new_with_label(_("OK"));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button, TRUE, TRUE, 0);
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            GTK_SIGNAL_FUNC(SaveDXFOK),
                            GTK_OBJECT(Dialog));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            GTK_SIGNAL_FUNC(gtk_widget_destroy),
                            GTK_OBJECT(Dialog));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(Button);
  gtk_widget_show(Button);

  /* The Cancel Button */
  Button = gtk_button_new_with_label(_("Cancel"));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(Dialog));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button, TRUE, TRUE, 0);
  gtk_widget_show(Button);
  
  /* The check button for enabling/disabling comment */
  CheckButton = gtk_check_button_new_with_label(_("Save with comment"));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), CheckButton, FALSE, FALSE, 2);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(CheckButton), TRUE);
  DXFData->IncludeComment = TRUE;
  gtk_signal_connect_object(GTK_OBJECT(CheckButton), "toggled",
                            GTK_SIGNAL_FUNC(ToggleDXFIncludeComment),
                            GTK_OBJECT(Dialog));
  gtk_widget_show(CheckButton);


  frame = gtk_frame_new("Comments:");
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_border_width(GTK_CONTAINER(frame), 10);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), frame, TRUE, TRUE, 0);
  gtk_widget_show(frame);

  Table = gtk_table_new(1, 1, FALSE);
  gtk_container_border_width(GTK_CONTAINER(Table), 10);
  gtk_container_add(GTK_CONTAINER(frame), Table);
  gtk_widget_show(Table);

  Text = gtk_text_new(NULL, NULL);
  gtk_text_set_editable(GTK_TEXT(Text), TRUE);
  gtk_text_insert(GTK_TEXT(Text), NULL, NULL, NULL, Comment, -1);
  gtk_table_attach(GTK_TABLE(Table), Text, 0, 1, 0, 1,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
  DXFData->Text = Text;
  gtk_widget_show(Text);

  /* Add a vertical scrollbar to the GtkText widget */
  Vscrollbar = gtk_vscrollbar_new(GTK_TEXT(Text)->vadj);
  gtk_table_attach(GTK_TABLE(Table), Vscrollbar, 1, 2, 0, 1,
                   GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
  gtk_widget_show(Vscrollbar);
 
  gtk_widget_show(Dialog);
  
}

/*****************************************************************************
*  SaveDXF
******************************************************************************/
void SaveDXF(GtkWidget *w, GtkFileSelection *FS)
{
  FrameStruct *LocalFrame;
  char *FileName;

  FileName = gtk_file_selection_get_filename(GTK_FILE_SELECTION(FS));
  LocalFrame = (FrameStruct *)gtk_object_get_data(GTK_OBJECT(FS), "frame");
  SaveDXFDialog(LocalFrame, FileName, FALSE);
}

/*****************************************************************************
*  SaveSelectedDXF
******************************************************************************/
void SaveSelectedDXF(GtkWidget *w, GtkFileSelection *FS)
{
  FrameStruct *LocalFrame;
  char *FileName;

  FileName = gtk_file_selection_get_filename(GTK_FILE_SELECTION(FS));
  LocalFrame = (FrameStruct *)gtk_object_get_data(GTK_OBJECT(FS), "frame");
  SaveDXFDialog(LocalFrame, FileName, TRUE);
}

