/* giram.h
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* This file should always be the first included. */
#include "config.h"
#include "giramlib/vectors.h"

#include <gtk/gtk.h>

/* OpenGL stuff */
#ifdef HAVE_GL
#include <gdk/gdkx.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#endif

gchar *AccelKeyFile;

#include "texture.h"

/* hash tables */
GHashTable *colors_hash;

typedef struct SkySphereStruct
{
  PigmentStruct **pPigmentList; /* Skysphere, is merely a list of layered
                                   pigments */
  int NbPigment;
  TransformStruct *Trans;
} SkySphereStruct;

/* What type of atmosphere is used in the frame? */
typedef struct AtmosphereStruct
{
  int Type;
  double Distance;
  double Scattering;
  double Eccentricity;
  int Samples;
  double Jitter;
  double AaThreshold;
  int AaLevel;
  Vector Color;                                      
} AtmosphereStruct;

/* Fog */
typedef struct FogStruct
{
  int FogType;
  double Distance;
  Vector Color;
  Vector Turbulence;
  double TurbDepth;
  double Omega;
  double Lambda;
  int Octaves;
  double FogOffset;
  double FogAlt;
  Vector Up;
  TransformStruct *Trans;
  struct FogStruct *Next;
} FogStruct;

/* The different types of camera. Only the PERSPECTIVE one is used. */
typedef enum
{
  CAMERA_PERSPECTIVE,
  CAMERA_ORTHOGRAPHIC,
  CAMERA_FISHEYE,
  CAMERA_ULTRA_WIDE_ANGLE,
  CAMERA_OMNIMAX,
  CAMERA_PANORAMIC,
  CAMERA_CYLINDER
} GiramCameraType;

typedef struct CameraStruct
{
  GiramCameraType Type;
  guint CylinderType;
  Vector Location;
  Vector LookAt;
  Vector Right;
  Vector Up;
  Vector Direction;
  Vector Sky;
  double Angle;
  int BlurSamples;
  double Aperture;
  Vector FocalPoint;
  double Confidence;
  double Variance;
  /* Private fields. These are used to calculate the Pers3d->2d */
  double x0, y0, z0;
  double PX, PY, PZ;
  double A, B, C, D;
  Vector P00, P01, P10, P11;
  Vector PA, PB, PC, PD;
  Vector Na, Nb, Nc;
  double Du0, Du1, Du2;
  double Dv0, Dv1, Dv2;
  Vector Qux, Quy, Qvx, Qvy;
  double Dux, Duy, Dvx, Dvy;
} CameraStruct;

/* A list of triangles is attached to each object to draw them on the screen */
typedef struct TriangleListStruct
{
  Vector P1, P2, P3;
  Vector N1, N2, N3;
  struct TriangleListStruct *Next;
} TriangleListStruct;

#include "object.h"

/* Here begins the differents objects definitions */
#define BOX_OBJECT 3
typedef struct BoxStruct
{
  ObjectStruct Object;

  Vector MinCorner;
  Vector MaxCorner;
} BoxStruct;

#define CONE_OBJECT 4
typedef struct ConeStruct
{
  ObjectStruct Object;

  int Open;
  double Radius1;
  double Radius2;
} ConeStruct;

typedef enum
{
  CSG_DIFFERENCE,
  CSG_INTERSECTION,
  CSG_MERGE,
  CSG_UNION
} GiramCSGType;

#define CSG_OBJECT 5
typedef struct CSGStruct
{
  ObjectStruct Object;

  GiramCSGType Type;
  GSList *all_objects;
} CSGStruct;

#define CYLINDER_OBJECT 7
typedef struct CylinderStruct
{
  ObjectStruct Object;

  int Open;
} CylinderStruct;

#define DISC_OBJECT 8
typedef struct DiscStruct
{
  ObjectStruct Object;

  double Radius, HoleRadius;
} DiscStruct;

#define MESH_OBJECT 12
typedef struct MeshStruct
{
  ObjectStruct Object;

  struct TriangleListStruct *FirstTriangle;
} MeshStruct;

#define PLANE_OBJECT 13
typedef struct PlaneStruct
{
  ObjectStruct Object;

  Vector Normal;
  double Distance;
  int NbSquare;
  double SquareSize;
} PlaneStruct;

#define SOR_OBJECT 19
typedef struct SorStruct
{
  ObjectStruct Object;

  guint NumberOfPoints;
  Vector *Point;
  gboolean Open;
} SorStruct;

#define SPHERE_OBJECT 20
typedef struct SphereStruct
{
  ObjectStruct Object;
  
  Vector Center;
  double Radius;
} SphereStruct;

#define SUPERELLIPSOID_OBJECT 21
typedef struct SuperEllipsoidStruct
{
  ObjectStruct Object;

  double E, N;
} SuperEllipsoidStruct;

#define TORUS_OBJECT 23
typedef struct TorusStruct
{
  ObjectStruct Object;

  double Major, Minor;
} TorusStruct;

#define TRIANGLE_OBJECT 24
typedef struct TriangleStruct
{
  ObjectStruct Object;

  Vector P1, P2, P3;
  Vector N1, N2, N3;
} TriangleStruct;

#define POINT_LIGHTSOURCE       0
#define SPOTLIGHT_LIGHTSOURCE   1
#define CYLINDER_LIGHTSOURCE    2
#define SHADOWLESS_LIGHTSOURCE 16
typedef struct LightSourceStruct
{
  Vector Location;
  Vector Color;
  int Type;
  Vector PointAt;
  double Radius;
  double FallOff;
  double Tightness;
  int AreaLight;
  Vector Area1, Area2;
  double AreaSize1, AreaSize2;
  double Adaptive;
  double Jitter;
  ObjectStruct *LooksLike;
  double FadeDistance;
  double FadePower;
  int AtmosphericAttenuation;
  int Atmosphere;
  struct LightSourceStruct *Next;
} LightSourceStruct;


/****************************************************************************
 * All the '*ConstantStruct' structures are there to hold the constant that *
 *  were created during the parsing of a Povray file.  I guess it would be  *
 *                       useful to use them elsewhere                       *
 ****************************************************************************/
typedef struct FloatConstantStruct
{
  double Float;
  char *Name;
  struct FloatConstantStruct *Next;
} FloatConstantStruct;

typedef struct VectorConstantStruct
{
  Vector Vector;
  char *Name;
  struct VectorConstantStruct *Next;
} VectorConstantStruct;

typedef struct ColorConstantStruct
{
  Vector Color;
  char *Name;
  struct ColorConstantStruct *Next;
} ColorConstantStruct;

typedef struct StringConstantStruct
{
  char *String;
  char *Name;
  struct StringConstantStruct *Next;
} StringConstantStruct;

typedef struct ObjectConstantStruct
{
  ObjectStruct *Object;
  char *Name;
  struct ObjectConstantStruct *Next;
} ObjectConstantStruct;

typedef struct TextureConstantStruct
{
  TextureStruct *Texture;
  char *Name;
  struct TextureConstantStruct *Next;
} TextureConstantStruct;

typedef struct PigmentConstantStruct
{
  PigmentStruct *Pigment;
  char *Name;
  struct PigmentConstantStruct *Next;
} PigmentConstantStruct;

typedef struct CameraConstantStruct
{
  CameraStruct *Camera;
  char *Name;
  struct CameraConstantStruct *Next;
} CameraConstantStruct;

typedef struct LightSourceConstantStruct
{
  LightSourceStruct *LightSource;
  char *Name;
  struct LightSourceConstantStruct *Next;
} LightSourceConstantStruct;

typedef struct FrameViewStruct
{
  gchar              *name;
  GtkWidget          *drawing_area;
  struct FrameStruct *Frame;
} FrameViewStruct;

/* ObjectSelection is a ringed linked list of the Objects that are
 * currently selected 
 */
typedef struct ObjectSelection
{
  ObjectStruct *Object;
  struct ObjectSelection *Next;
} ObjectSelection;

typedef struct FrameStruct
{
  char *Name; /* Its name */
  int Id;     /* This is used by the plugins interface */
  CameraStruct *Camera;
/*  ObjectStruct *FirstObject;*/
  GSList *all_objects;
  LightSourceStruct *FirstLightSource;
  SkySphereStruct *SkySphere;
  /* All the constants linked list... */
  FloatConstantStruct *FloatConstantList;
  VectorConstantStruct *VectorConstantList;
  ColorConstantStruct *ColorConstantList;
  StringConstantStruct *StringConstantList;
  ObjectConstantStruct *ObjectConstantList;
  TextureConstantStruct *TextureConstantList;
  PigmentConstantStruct *PigmentConstantList;
  CameraConstantStruct *CameraConstantList;
  LightSourceConstantStruct *LightSourceConstantList;
  
  GSList *all_views;
  ObjectSelection *Selection;
  Vector RotationCenter;
  Vector ScaleCenter;
  /* The things for the snap... */
  gboolean SnapToGrid, ShowGrid;
  Vector   SnapGrid;
  struct FrameStruct *Next; /* The Next frame in Giram */
} FrameStruct;

FrameStruct *FirstFrame;

#define FRAME_CAMERA 1
#define ORTHO_XY_CAMERA 2
#define ORTHO_XZ_CAMERA 3
#define ORTHO_ZY_CAMERA 4

/* This is attached to every view, so we know how to render them */
typedef struct ViewUserDataStruct
{
  GtkWidget *DrawingArea;
  int CameraStyle;
  FrameStruct *FrameView;
  double ZoomFactor;
  double Xoffset, Yoffset;
  guint RenderStyle;
  GtkWidget *StatusBar;
#ifdef HAVE_GL
  GLXContext glxcontext;
#endif
} ViewUserDataStruct;

extern ViewUserDataStruct *CurrentUserData;

#ifdef HAVE_GL
GdkVisual *GLVisual;
int ReallyHaveGL;
#endif

/* This is used for the Gouraud rendering funcs */
typedef struct Triangle2D
{
  Vector P1,P2,P3;
  double distance;
  GdkColor Color;
} Triangle2D;

typedef struct Triangle2DNormCol
{
  int P[3][2];
  double Color[3][3];
  double Distance[3];
} Triangle2DNormCol;

/* The tools of the main tool box. */
typedef enum
{
  MT_SELECT                 =  1,
  MT_MOVE                   =  2,
  MT_ROTATE                 =  3,
  MT_SCALE                  =  4,
  MT_NEW_BOX                =  5,
  MT_NEW_CONE               =  6,
  MT_NEW_CYLINDER           =  7,
  MT_NEW_DISC               =  8,
  MT_NEW_PLANE              =  9,
  MT_NEW_SOR                = 10,
  MT_NEW_SPHERE             = 11,
  MT_NEW_SUPERELLIPSOID     = 12,
  MT_NEW_TORUS              = 13,
  MT_NEW_TRIANGLE           = 14,
  MT_NEW_LIGHTSOURCE        = 15,
  MT_LAST
} GiramModeToolType;

extern GiramModeToolType ModeTool;

/* The prototypes used in giram.c */
void CreateView(void);
void LoadPovFunc(void);
void LoadDXFFunc(void);
#ifdef USE_LIB3DS
void Load3DSFunc(void);
void Insert3DSFunc(void);
#endif
void InsertDXFFunc(void);
void SavePovFunc(void);
void SaveDXFFunc(void);
void SaveSelectedDXFFunc(void);
#ifdef WITH_S3D
void SaveS3DFunc(void);
void SaveSelectedS3DFunc(void);
#endif
void ExitFunc(void);
int comparTri(const void *Tri1, const void *Tri2);
void DrawGouraudTriangleZBuffer(GtkWidget *View, double *Zbuffer,
                                    Triangle2DNormCol *Tri, guchar *rgb_buf);

gchar *help_path, *plugins_path, *temp_path;
gint toolbox_style;
gint dynamic_toolbar_shape;
