/* gtkcolorbutton.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This file is heavily inspired from the same one from the Gimp:
 * Gimp Color Button
 * Copyright (C) 1999 Sven Neumann
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU  
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "giram.h"

#include "gtkcolorbutton.h"

#include "giramcolorlist.h"

static void  gtk_color_button_destroy(GtkObject *object);
static void  gtk_color_button_clicked(GtkButton *button);
static void  gtk_color_button_paint(GtkColorButton *gcb);
static void  gtk_color_button_state_changed(GtkWidget *widget, GtkStateType previous_state);
static void  gtk_color_button_color_changed(GtkColorButton *gcb, gdouble *color);

enum
{
  COLOR_CHANGED,
  LAST_SIGNAL
};

static guint gtk_color_button_signals[LAST_SIGNAL] = { 0 };


static GtkWidgetClass *parent_class = NULL;

/*************************************************************************
*
**************************************************************************/
static void gtk_color_button_destroy(GtkObject *object)
{
  GtkColorButton *gcb;
   
  g_return_if_fail(gcb = GTK_COLOR_BUTTON(object));
  
  g_free(gcb->title);

  if (gcb->gcl)
    gtk_widget_destroy(gcb->gcl);

  g_free(gcb->color);

  if (GTK_OBJECT_CLASS(parent_class)->destroy)
    (* GTK_OBJECT_CLASS(parent_class)->destroy)(object);
}

/*************************************************************************
*
**************************************************************************/
static void gtk_color_button_class_init(GtkColorButtonClass *class)
{
  GtkObjectClass *object_class;
  GtkButtonClass *button_class;
  GtkWidgetClass *widget_class;

  object_class = (GtkObjectClass*) class;
  button_class = (GtkButtonClass*) class;
  widget_class = (GtkWidgetClass*) class;

  parent_class = gtk_type_class(gtk_widget_get_type());

  gtk_color_button_signals[COLOR_CHANGED] = 
          gtk_signal_new("color_changed",
                         GTK_RUN_FIRST,
                         object_class->type,
                         GTK_SIGNAL_OFFSET(GtkColorButtonClass, color_changed),
                         gtk_signal_default_marshaller, GTK_TYPE_NONE, 0);
  gtk_object_class_add_signals(object_class, gtk_color_button_signals, 
                               LAST_SIGNAL);
  class->color_changed = NULL;

  object_class->destroy       = gtk_color_button_destroy;
  button_class->clicked       = gtk_color_button_clicked;
  widget_class->state_changed = gtk_color_button_state_changed;
}

/*************************************************************************
*
**************************************************************************/
static void gtk_color_button_init(GtkColorButton *gcb)
{
  gcb->title    = NULL;

  gcb->color    = NULL;
  gcb->preview  = NULL;
  gcb->gcl      = NULL;
}

/*************************************************************************
*
**************************************************************************/
GtkType gtk_color_button_get_type(void)
{
  static guint gcb_type = 0;

  if (!gcb_type)
    {
      GtkTypeInfo gcb_info =
      {
        "GtkColorButton",
        sizeof(GtkColorButton),
        sizeof(GtkColorButtonClass),
        (GtkClassInitFunc)gtk_color_button_class_init,
        (GtkObjectInitFunc)gtk_color_button_init,
        /* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL
      };

      gcb_type = gtk_type_unique(gtk_button_get_type(), &gcb_info);
    }

  return gcb_type;
}
/**
 * gtk_color_button_new:
 * @title: String that wil be used as title for the color_selector.
 * @width: Width of the colorpreview in pixels.
 * @height: Height of the colorpreview in pixels.
 * @color: An array of guchar holding the color (RGB or RGBA)
 * @bpp: May be 3 for RGB or 4 for RGBA.
 * 
 * Creates a new GtkColorbutton widget. This returns a button with 
 * a preview showing the color. When the button is clicked a 
 * GtkColorSelectionDialog is opened. If the user changes the color
 * the new color is written into the array that was used to pass
 * the initial color and the "color_changed" signal is emitted.
 * 
 * Returns: Pointer to the new GtkColorButton widget.
 */

GtkWidget *gtk_color_button_new(gchar *title,
                                gint width, gint height,
                                gdouble *color)
{
  GtkColorButton *gcb;
  gint i;
  
  g_return_val_if_fail(width > 0 && height > 0, NULL);  

  gcb = gtk_type_new(gtk_color_button_get_type());

  gcb->title  = g_strdup(title);
  gcb->width  = width;
  gcb->height = height;

  gcb->color = g_new(gdouble, 5);

  for (i = 0 ; i < 5 ; i++)
    gcb->color[i] = color[i];

  gcb->preview = gtk_preview_new(GTK_PREVIEW_COLOR);
  gtk_signal_connect(GTK_OBJECT(gcb->preview), "destroy",
                     gtk_widget_destroyed, &gcb->preview);

  gtk_preview_size(GTK_PREVIEW(gcb->preview), width, height);
  gtk_container_add(GTK_CONTAINER(gcb), gcb->preview);
  gtk_widget_show(gcb->preview);
  gtk_color_button_paint(gcb);

  return (GTK_WIDGET(gcb));
}

/*
 * gtk_color_button_update:
 * @gcb: Pointer to a #GimpColorButton.
 * 
 * Should be used after the color controlled by a #GimpColorButton
 * was changed. The color is then reread and the change is propagated
 * to the preview and the GtkColorSelectionDialog if one is open.
 */
void gtk_color_button_update(GtkColorButton *gcb)
{
  gint i;

  g_return_if_fail(GTK_IS_COLOR_BUTTON(gcb));

  gtk_color_button_paint(gcb);

  if (gcb->gcl)
    giram_color_list_set_color(GIRAM_COLOR_LIST(gcb->gcl), gcb->color);
}

static void gtk_color_button_state_changed(GtkWidget *widget,
                                           GtkStateType previous_state)
{  
  g_return_if_fail (GTK_IS_COLOR_BUTTON (widget));

  if (!GTK_WIDGET_IS_SENSITIVE(widget) && GTK_COLOR_BUTTON(widget)->gcl)
    gtk_widget_hide(GTK_COLOR_BUTTON(widget)->gcl);

  if (GTK_WIDGET_CLASS(parent_class)->state_changed)
    (* GTK_WIDGET_CLASS(parent_class)->state_changed)(widget, previous_state);
}

static void gtk_color_button_clicked(GtkButton *button)
{
  GtkColorButton *gcb;
  GtkColorSelection *colorsel;

  g_return_if_fail (GTK_IS_COLOR_BUTTON(button));
  gcb = GTK_COLOR_BUTTON(button);

  if (!gcb->gcl)
  {
    gcb->gcl = giram_color_list_new(colors_hash);
    gtk_signal_connect(GTK_OBJECT(gcb->gcl), "destroy",
                       (GtkSignalFunc)gtk_widget_destroyed, &gcb->gcl);
    gtk_signal_connect_object(GTK_OBJECT(gcb->gcl), "color_changed",
                              (GtkSignalFunc)gtk_color_button_color_changed,
                              GTK_OBJECT(gcb));
    gtk_window_position(GTK_WINDOW(gcb->gcl), GTK_WIN_POS_MOUSE);
  }
  giram_color_list_set_color(GIRAM_COLOR_LIST(gcb->gcl), gcb->color);
  gtk_widget_show(gcb->gcl);
}

static void gtk_color_button_paint(GtkColorButton *gcb)
{
  gint x, y, i;
  guchar *p0, *row;

  g_return_if_fail(GTK_IS_COLOR_BUTTON(gcb));

  p0 = row = g_new(guchar, 3*gcb->width);
  for (x = 0; x < gcb->width; x++)
  {
    for (i = 0; i < 3; i++)
      *p0++ = gcb->color[i] * 255.0;
  }
  for (y = 0; y < gcb->height; y++)
    gtk_preview_draw_row(GTK_PREVIEW(gcb->preview), row, 0, y,
                         gcb->width);
  gtk_widget_queue_draw (gcb->preview);
  g_free(row);
}

static void gtk_color_button_color_changed(GtkColorButton *gcb, gdouble *color)
{
  gint i;

  for (i=0 ; i<5 ; i++)
    gcb->color[i] = color[i];
  gtk_color_button_paint(gcb);
  gtk_signal_emit(GTK_OBJECT(gcb),  gtk_color_button_signals[COLOR_CHANGED]);
}
