/* object.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "giram.h"
#include "csg.h"
#include "mesh.h"
#include "plane.h"
#include "sor.h"
#include "sphere.h"
#include "torus.h"
#include "triangle.h"
#include "copy.h"

static void giram_object_translate_real(ObjectStruct *object, Vector vect);
static void giram_object_rotate_real(ObjectStruct *object, Vector vect);
static void giram_object_scale_real(ObjectStruct *object, Vector vect);
static void giram_object_build_triangle_mesh_real(ObjectStruct *object);
static gboolean giram_object_inside_real(ObjectStruct *object, double x, double y, double z);
static gboolean giram_object_is_intersection_real(ObjectStruct *object, Vector origin, Vector direction);
static gboolean giram_object_click_in_XY_real(ObjectStruct *object, double x, double y);
static gboolean giram_object_click_in_XZ_real(ObjectStruct *object, double x, double z);
static gboolean giram_object_click_in_ZY_real(ObjectStruct *object, double z, double y);
static gboolean giram_object_find_intersection_segment_real(ObjectStruct *object,
                                                            Vector in_point, Vector out_point,
                                                            Vector inter_point, Vector inter_norm);

/*****************************************************************************
*  InitObject
******************************************************************************/
void InitObject(ObjectStruct *Object)
{
  static Id = 1;

  Object->Name = NULL;
  Object->FirstTriangle = NULL;
  Object->Trans = NULL;
  Object->Texture = CopyTexture(DefaultTexture);
  Object->Inverse = 0;
  Object->NoShadow = 0;
  Object->Hollow = 0;
  Object->Selected = 0;
  Object->Visible = 1;
  Object->CSGTreeNode = NULL;
  Object->Id = Id++;
}

/* FIXME: Should be OOized */

/*****************************************************************************
*  giram_object_class_new
******************************************************************************/
GiramObjectClass *giram_object_class_new(void)
{
  GiramObjectClass *object_class;

  object_class = g_new(GiramObjectClass, 1);
  object_class->translate                 = giram_object_translate_real;
  object_class->rotate                    = giram_object_rotate_real;
  object_class->scale                     = giram_object_scale_real;
  object_class->build_triangle_mesh       = giram_object_build_triangle_mesh_real;
  object_class->inside                    = giram_object_inside_real;
  object_class->is_intersection           = giram_object_is_intersection_real;
  object_class->click_in_XY               = giram_object_click_in_XY_real;
  object_class->click_in_XZ               = giram_object_click_in_XZ_real;
  object_class->click_in_ZY               = giram_object_click_in_ZY_real;
  object_class->find_intersection_segment = giram_object_find_intersection_segment_real;
  return object_class;
}

/*****************************************************************************
*  giram_object_translate
******************************************************************************/
void giram_object_translate(ObjectStruct *object, Vector vect)
{
  object->klass->translate(object, vect);
}

/*****************************************************************************
*  giram_object_translate_real
******************************************************************************/
static void giram_object_translate_real(ObjectStruct *object, Vector vect)
{
  TransformStruct Trans;

  if (object->Trans)
  {
    ComputeTranslateTrans(&Trans, vect);
    ComposeTrans(object->Trans, &Trans);
  } else
  {
    object->Trans = g_malloc(sizeof(TransformStruct));
    ComputeTranslateTrans(object->Trans, vect);
  }
}

/*****************************************************************************
*  giram_object_rotate
******************************************************************************/
void giram_object_rotate(ObjectStruct *object, Vector vect)
{
  object->klass->rotate(object, vect);
}

/*****************************************************************************
*  giram_object_rotate_real
******************************************************************************/
static void giram_object_rotate_real(ObjectStruct *object, Vector vect)
{
  TransformStruct Trans;

  if (object->Trans)
  {
    ComputeRotateTrans(&Trans, vect);
    ComposeTrans(object->Trans, &Trans);
  } else
  {
    object->Trans = g_malloc(sizeof(TransformStruct));
    ComputeRotateTrans(object->Trans, vect);
  }
}

/*****************************************************************************
*  giram_object_scale
******************************************************************************/
void giram_object_scale(ObjectStruct *object, Vector vect)
{
  object->klass->scale(object, vect);
}

/*****************************************************************************
*  giram_object_scale_real
******************************************************************************/
static void giram_object_scale_real(ObjectStruct *object, Vector vect)
{
  TransformStruct Trans;

  if (object->Trans)
  {
    ComputeScaleTrans(&Trans, vect);
    ComposeTrans(object->Trans, &Trans);
  } else
  {
    object->Trans = g_malloc(sizeof(TransformStruct));
    ComputeScaleTrans(object->Trans, vect);
  }
}

/*****************************************************************************
*  giram_object_build_triangle_mesh
******************************************************************************/
void giram_object_build_triangle_mesh(ObjectStruct *object)
{
  object->klass->build_triangle_mesh(object);
}

/*****************************************************************************
*  giram_object_build_triangle_mesh_real
******************************************************************************/
static void giram_object_build_triangle_mesh_real(ObjectStruct *object)
{
  object->FirstTriangle = NULL;
}

/*****************************************************************************
*  giram_object_inside
******************************************************************************/
gboolean giram_object_inside(ObjectStruct *object, double x, double y, double z)
{
  return object->klass->inside(object, x, y, z);
}

/*****************************************************************************
*  giram_object_inside_real
******************************************************************************/
static gboolean giram_object_inside_real(ObjectStruct *object, double x, double y, double z)
{
  return FALSE;
}

/*****************************************************************************
*  giram_object_is_intersection
******************************************************************************/
gboolean giram_object_is_intersection(ObjectStruct *object, Vector orig, Vector dir)
{
  return object->klass->is_intersection(object, orig, dir);
}

/*****************************************************************************
*  giram_object_is_intersection_real
******************************************************************************/
static gboolean giram_object_is_intersection_real(ObjectStruct *object, Vector origin, Vector direction)
{
  Vector org, dir;
  int i;
  Matrix Mat, InvMat;
  double D;
  double u, v;
  TriangleListStruct *TmpTri;

  if (object->Trans)
  {
    MInverseEvaluatePoint(org, object->Trans, origin);
    MInverseEvaluateVector(dir, object->Trans, direction);
  } else
  {
    for (i=0 ; i<3 ; i++)
    {
      org[i] = origin[i];
      dir[i] = direction[i];
    }
  }
  for (TmpTri = object->FirstTriangle ; TmpTri ; TmpTri = TmpTri->Next)
  {
    Mat[0][0] = TmpTri->P2[0] - TmpTri->P1[0];
    Mat[0][1] = TmpTri->P2[1] - TmpTri->P1[1];
    Mat[0][2] = TmpTri->P2[2] - TmpTri->P1[2];
    Mat[1][0] = TmpTri->P3[0] - TmpTri->P1[0];
    Mat[1][1] = TmpTri->P3[1] - TmpTri->P1[1];
    Mat[1][2] = TmpTri->P3[2] - TmpTri->P1[2];
    Mat[2][0] = -dir[0];
    Mat[2][1] = -dir[1];
    Mat[2][2] = -dir[2];
    D = Mat[0][0]*Mat[1][1]*Mat[2][2]+Mat[0][1]*Mat[1][2]*Mat[2][0]+
        Mat[0][2]*Mat[1][0]*Mat[2][1]-Mat[2][0]*Mat[1][1]*Mat[0][2]-
        Mat[2][1]*Mat[1][2]*Mat[0][0]-Mat[2][2]*Mat[1][0]*Mat[0][1];
    if (D != 0.0)
    {
      InvMat[0][0] =  (Mat[1][1]*Mat[2][2]-Mat[2][1]*Mat[1][2])/D;
      InvMat[0][1] = -(Mat[0][1]*Mat[2][2]-Mat[2][1]*Mat[0][2])/D;
      InvMat[0][2] =  (Mat[0][1]*Mat[1][2]-Mat[1][1]*Mat[0][2])/D;
      InvMat[1][0] = -(Mat[1][0]*Mat[2][2]-Mat[2][0]*Mat[1][2])/D;
      InvMat[1][1] =  (Mat[0][0]*Mat[2][2]-Mat[2][0]*Mat[0][2])/D;
      InvMat[1][2] = -(Mat[0][0]*Mat[1][2]-Mat[1][0]*Mat[0][2])/D;
      InvMat[2][0] =  (Mat[1][0]*Mat[2][1]-Mat[2][0]*Mat[1][1])/D;
      InvMat[2][1] = -(Mat[0][0]*Mat[2][1]-Mat[2][0]*Mat[0][1])/D;
      InvMat[2][2] =  (Mat[0][0]*Mat[1][1]-Mat[1][0]*Mat[0][1])/D;
      u = InvMat[0][0]*(org[0]-TmpTri->P1[0])+
          InvMat[1][0]*(org[1]-TmpTri->P1[1])+
          InvMat[2][0]*(org[2]-TmpTri->P1[2]);
      v = InvMat[0][1]*(org[0]-TmpTri->P1[0])+
          InvMat[1][1]*(org[1]-TmpTri->P1[1])+
          InvMat[2][1]*(org[2]-TmpTri->P1[2]);
      if ((u>=0.0) && (u<=1.0) && (v>=0.0) && (v<=1.0) && (u+v<=1.0))
        return TRUE;
    }
  }
  return FALSE;
}

/*****************************************************************************
*  giram_object_click_in_XY
******************************************************************************/
gboolean giram_object_click_in_XY(ObjectStruct *object, double x, double y)
{
  return object->klass->click_in_XY(object, x, y);
}

/*****************************************************************************
*  giram_object_click_in_XY_real
******************************************************************************/
static gboolean giram_object_click_in_XY_real(ObjectStruct *object, double x, double y)
{
  Vector Origin = {0.0, 0.0, 0.0, 0.0, 0.0};
  Vector Direction = {0.0, 0.0, 1.0, 0.0, 0.0};

  Origin[0] = x;
  Origin[1] = y;
  return giram_object_is_intersection(object, Origin, Direction);
}

/*****************************************************************************
*  giram_object_click_in_XZ
******************************************************************************/
gboolean giram_object_click_in_XZ(ObjectStruct *object, double x, double z)
{
  return object->klass->click_in_XZ(object, x, z);
}

/*****************************************************************************
*  giram_object_click_in_XZ_real
******************************************************************************/
static gboolean giram_object_click_in_XZ_real(ObjectStruct *object, double x, double z)
{
  Vector Origin = {0.0, 0.0, 0.0, 0.0, 0.0};
  Vector Direction = {1.0, 0.0, 0.0, 0.0, 0.0};

  Origin[0] = x;
  Origin[2] = z;
  return giram_object_is_intersection(object, Origin, Direction);
}

/*****************************************************************************
*  giram_object_click_in_ZY
******************************************************************************/
gboolean giram_object_click_in_ZY(ObjectStruct *object, double z, double y)
{
  return object->klass->click_in_ZY(object, z, y);
}

/*****************************************************************************
*  giram_object_click_in_ZY_real
******************************************************************************/
static gboolean giram_object_click_in_ZY_real(ObjectStruct *object, double z, double y)
{
  Vector Origin = {0.0, 0.0, 0.0, 0.0, 0.0};
  Vector Direction = {1.0, 0.0, 0.0, 0.0, 0.0};

  Origin[2] = z;
  Origin[1] = y;
  return giram_object_is_intersection(object, Origin, Direction);
}

/*****************************************************************************
*  giram_object_find_intersection_segment
******************************************************************************/
gboolean giram_object_find_intersection_segment(ObjectStruct *object,
                                                Vector in_point, Vector out_point,
                                                Vector inter_point, Vector inter_norm)
{
  return object->klass->find_intersection_segment(object, in_point, out_point,
                                                          inter_point, inter_norm);
}

/*****************************************************************************
*  giram_object_find_intersection_segment_real
******************************************************************************/
static gboolean giram_object_find_intersection_segment_real(ObjectStruct *object,
                                                            Vector in_point,
                                                            Vector out_point,
                                                            Vector inter_point,
                                                            Vector inter_norm)
{
  return FALSE;
}

/*****************************************************************************
*  PDBTransformObject
******************************************************************************/
int PDBTransformObject(int *FrameId, guint *ObjectId, TransformStruct *Trans)
{
  FrameStruct *TmpFrame;
  ObjectStruct *TmpObject;
  GSList *tmp_list;
  int i,j;
  
  for (TmpFrame = FirstFrame ;
       TmpFrame->Id != *FrameId ;
       TmpFrame = TmpFrame->Next)
     ;
  for (tmp_list = TmpFrame->all_objects ;
       tmp_list && tmp_list->data && ((ObjectStruct*)(tmp_list->data))->Id != *ObjectId ;
       tmp_list = g_slist_next(tmp_list))
     ;
  TmpObject = (ObjectStruct *)(tmp_list->data);
  if (TmpObject->Trans)
  {
    ComposeTrans(TmpObject->Trans, Trans);
  } else
  {
    TmpObject->Trans = g_malloc(sizeof(TransformStruct));
    for (i=0 ; i<4 ; i++)
      for (j=0 ; j<4 ; j++)
      {
        TmpObject->Trans->Direct[i][j] = Trans->Direct[i][j];
        TmpObject->Trans->Inverse[i][j] = Trans->Inverse[i][j];
      }
  }
  giram_object_build_triangle_mesh(TmpObject);
  return *ObjectId;
}

