/* plane.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "giram.h"
#include "trimesh.h"
#include "csgtree.h"
#include "object.h"
#include "utils.h"
#include "plane.h"

static void giram_plane_translate(ObjectStruct *plane, Vector vect);
static void giram_plane_rotate(ObjectStruct *plane, Vector vect);
static void giram_plane_scale(ObjectStruct *plane, Vector vect);
static void giram_plane_build_triangle_mesh(ObjectStruct *plane);
static gboolean giram_plane_inside(ObjectStruct *plane, double x, double y, double z);
static gboolean giram_plane_find_intersection_segment(ObjectStruct *plane,
                                                      Vector in_point, Vector out_point,
                                                      Vector inter_point, Vector inter_norm);

/*****************************************************************************
*  giram_plane_translate
******************************************************************************/
static void giram_plane_translate(ObjectStruct *plane, Vector vect)
{
  TransformStruct Trans;
  PlaneStruct *pplane = (PlaneStruct *)plane;

  if (plane->Trans)
  {
    ComputeTranslateTrans(&Trans, vect);
    ComposeTrans(plane->Trans, &Trans);
  } else
  {
    pplane->Distance -= VDot(pplane->Normal, vect); 
  }
}

/*****************************************************************************
*  giram_plane_rotate
******************************************************************************/
static void giram_plane_rotate(ObjectStruct *plane, Vector vect)
{
  TransformStruct Trans;
  PlaneStruct *pplane = (PlaneStruct *)plane;

  ComputeRotateTrans(&Trans, vect);
  if (plane->Trans)
  {
    ComposeTrans(plane->Trans, &Trans);
  } else
  {
    MEvaluateVector(pplane->Normal, &Trans, pplane->Normal); 
  }
}

/*****************************************************************************
*  giram_plane_scale
******************************************************************************/
static void giram_plane_scale(ObjectStruct *plane, Vector vect)
{
  TransformStruct Trans;
  double Length;
  PlaneStruct *pplane = (PlaneStruct *)plane;

  if (plane->Trans)
  {
    ComputeScaleTrans(&Trans, vect);
    ComposeTrans(plane->Trans, &Trans);
  } else
  {
    pplane->Normal[0] /= vect[0];
    pplane->Normal[1] /= vect[1];
    pplane->Normal[2] /= vect[2];
    Length = VLength(pplane->Normal);
    pplane->Normal[0] /= Length;
    pplane->Normal[1] /= Length;
    pplane->Normal[2] /= Length;
    pplane->Distance  /= Length;
  }
}

/*****************************************************************************
*  giram_plane_build_triangle_mesh
******************************************************************************/
static void giram_plane_build_triangle_mesh(ObjectStruct *plane)
{ /* XXX: clipped_by */
  Vector CentralPoint;
  Vector U1, U2;
  double ULength;
  Vector P1, P2, P3, P4;
  Vector N;
  int i,j;
  PlaneStruct *pplane = (PlaneStruct *)plane;

  if (plane->FirstTriangle)
    DestroyObjectTriangleMesh(plane);
  N[0] = pplane->Normal[0];
  N[1] = pplane->Normal[1];
  N[2] = pplane->Normal[2];
  if (plane->Inverse)
  {
    N[0] = -N[0];
    N[1] = -N[1];
    N[2] = -N[2];
  }
  CentralPoint[0] = pplane->Normal[0]*pplane->Distance;
  CentralPoint[1] = pplane->Normal[1]*pplane->Distance;
  CentralPoint[2] = pplane->Normal[2]*pplane->Distance;

  if ( (fabs(N[0])>fabs(N[1])) && (fabs(N[0])>fabs(N[2])) )
  { /* The plane normal is upon the X axis */
    U1[0] = -N[1]/N[0];
    U1[1] = 1;
    U1[2] = 0;
  } else if ( (fabs(N[1])>fabs(N[0])) && (fabs(N[1])>fabs(N[2])) )
  { /* The plane normal is upon the Y axis */
    U1[0] = 1;
    U1[1] = -N[0]/N[1];
    U1[2] = 0;
  } else
  { /* The plane normal is upon the Z axis */
    U1[0] = 0;
    U1[1] = 1;
    U1[2] = -N[1]/N[2];
  }
  ULength = VLength(U1);
  U1[0]/=ULength; U1[1]/=ULength; U1[2]/=ULength;
  VCross(U2, N, U1);
  ULength = VLength(U2);
  U2[0]/=ULength; U2[1]/=ULength; U2[2]/=ULength;
  
  if (plane->Trans)
  {
    MEvaluateVector(N, plane->Trans, N);
  }
  
  for (i=-pplane->NbSquare; i<pplane->NbSquare ;i++)
    for (j=-pplane->NbSquare; j<pplane->NbSquare ;j++)
    {
      P1[0] = CentralPoint[0]+pplane->SquareSize*(i*U1[0]+j*U2[0]);
      P1[1] = CentralPoint[1]+pplane->SquareSize*(i*U1[1]+j*U2[1]);
      P1[2] = CentralPoint[2]+pplane->SquareSize*(i*U1[2]+j*U2[2]);

      P2[0] = CentralPoint[0]+pplane->SquareSize*((i+1)*U1[0]+j*U2[0]);
      P2[1] = CentralPoint[1]+pplane->SquareSize*((i+1)*U1[1]+j*U2[1]);
      P2[2] = CentralPoint[2]+pplane->SquareSize*((i+1)*U1[2]+j*U2[2]);

      P3[0] = CentralPoint[0]+pplane->SquareSize*(i*U1[0]+(j+1)*U2[0]);
      P3[1] = CentralPoint[1]+pplane->SquareSize*(i*U1[1]+(j+1)*U2[1]);
      P3[2] = CentralPoint[2]+pplane->SquareSize*(i*U1[2]+(j+1)*U2[2]);

      P4[0] = CentralPoint[0]+pplane->SquareSize*((i+1)*U1[0]+(j+1)*U2[0]);
      P4[1] = CentralPoint[1]+pplane->SquareSize*((i+1)*U1[1]+(j+1)*U2[1]);
      P4[2] = CentralPoint[2]+pplane->SquareSize*((i+1)*U1[2]+(j+1)*U2[2]);

      if (plane->Trans)
      {
        MEvaluatePoint(P1, plane->Trans, P1);
        MEvaluatePoint(P2, plane->Trans, P2);
        MEvaluatePoint(P3, plane->Trans, P3);
        MEvaluatePoint(P4, plane->Trans, P4);
      }
      AddTriangleToObjectMesh(plane, P1, P2, P3, N,N,N);
      AddTriangleToObjectMesh(plane, P2, P3, P4, N,N,N);
    }
}

/*****************************************************************************
*  giram_plane_inside
******************************************************************************/
static gboolean giram_plane_inside(ObjectStruct *plane, double x, double y, double z)
{ /* XXX: clipped_by */
  double a;
  Vector Vect;
  PlaneStruct *pplane = (PlaneStruct *)plane;

  Vect[0] = x;
  Vect[1] = y;
  Vect[2] = z;
  if (plane->Trans)
  {
    MInverseEvaluatePoint(Vect, plane->Trans, Vect);
  }
  a  = Vect[0]*pplane->Normal[0];
  a += Vect[1]*pplane->Normal[1];
  a += Vect[2]*pplane->Normal[2];
  if (a > pplane->Distance)
    return plane->Inverse;
  else
    return !(plane->Inverse);
}

/*****************************************************************************
*  giram_plane_find_intersection_segment
******************************************************************************/
static gboolean giram_plane_find_intersection_segment(ObjectStruct *plane,
                                                      Vector in_point, Vector out_point,
                                                      Vector inter_point, Vector inter_norm)
{
  return FALSE;
}

/*****************************************************************************
*  ToolNewPlane
******************************************************************************/
void ToolNewPlane(GtkWidget *DrawingArea, GdkEventButton *bevent)
{
  ObjectStruct       *Plane;
  Vector              Normal;
  ViewUserDataStruct *UserData;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;
  
  UserData = get_view_user_data();
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      Normal[0] = 0.0;
      Normal[1] = 0.0;
      Normal[2] = 1.0;
      break;
    case ORTHO_XZ_CAMERA:
      Normal[0] = 0.0;
      Normal[1] = 1.0;
      Normal[2] = 0.0;
      break;
    case ORTHO_ZY_CAMERA:
      Normal[0] = 1.0;
      Normal[1] = 0.0;
      Normal[2] = 0.0;
      break;
  }
  Plane = giram_plane_new(Normal, 0.0);
  UserData->FrameView->all_objects = g_slist_append(UserData->FrameView->all_objects,
                                                    Plane);
  giram_object_build_triangle_mesh(Plane);
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, UserData->FrameView);
}

/*****************************************************************************
*  giram_plane_new
******************************************************************************/
ObjectStruct *giram_plane_new(Vector Normal, double Distance)
{
  ObjectStruct *Plane;
  PlaneStruct *PPlane;
  static GiramObjectClass *plane_klass = NULL;

  if (plane_klass == NULL)
  {
    plane_klass = giram_object_class_new();

    plane_klass->translate                 = giram_plane_translate;
    plane_klass->rotate                    = giram_plane_rotate;
    plane_klass->scale                     = giram_plane_scale;
    plane_klass->build_triangle_mesh       = giram_plane_build_triangle_mesh;
    plane_klass->inside                    = giram_plane_inside;
    plane_klass->find_intersection_segment = giram_plane_find_intersection_segment;
  }
  PPlane = g_new(PlaneStruct, 1);
  Plane = (ObjectStruct *)PPlane;
  InitObject(Plane);
  Plane->Type = PLANE_OBJECT;
  Plane->klass = plane_klass;
  PPlane->Normal[0] = Normal[0];
  PPlane->Normal[1] = Normal[1];
  PPlane->Normal[2] = Normal[2];
  PPlane->Distance = Distance;
  PPlane->NbSquare = 4;
  PPlane->SquareSize = 2.5;
  return Plane;
}


/*****************************************************************************
*  PDBCreatePlane
******************************************************************************/
int PDBCreatePlane(int *Id, Vector *Normal, double *Distance)
{
  ObjectStruct    *Plane;
  PlaneStruct     *PPlane;
  FrameStruct     *LocalFrame;
  GSList          *tmp_list;
  FrameViewStruct *TmpView;

  Plane = giram_plane_new(*Normal, *Distance);

  for (LocalFrame = FirstFrame ;
       LocalFrame->Id != *Id ;
       LocalFrame = LocalFrame->Next)
                  ;
  LocalFrame->all_objects = g_slist_append(LocalFrame->all_objects, Plane);
  giram_object_build_triangle_mesh(Plane);
  for (tmp_list = LocalFrame->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, LocalFrame);
  return Plane->Id;
}

