/* torus.c
 * Giram - A GPLed Modelling Program.
 * Copyright (C) 1999-2000 DindinX <David.Odin@bigfoot.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include "giram.h"
#include "trimesh.h"
#include "csgtree.h"
#include "object.h"
#include "snap.h"
#include "utils.h"
#include "root_solver.h"
#include "torus.h"

#include "giramintl.h"

static void giram_torus_build_triangle_mesh(ObjectStruct *torus);
static gboolean giram_torus_inside(ObjectStruct *torus, double x, double y, double z);
static gboolean giram_torus_is_intersection(ObjectStruct *torus, Vector org, Vector dir);
static gboolean giram_torus_find_intersection_segment(ObjectStruct *torus,
                                                      Vector in_point, Vector out_point,
                                                      Vector inter_point, Vector inter_norm);

/*****************************************************************************
*  giram_torus_build_triangle_mesh
******************************************************************************/
static void giram_torus_build_triangle_mesh(ObjectStruct *torus)
{ /* XXX: clipped_by */
  int i,j;
  int DetailLevel;
  Vector P1, P2, P3, P4;
  Vector N1, N2, N3, N4;
  TorusStruct *ttorus = (TorusStruct *)torus;

  DetailLevel = 16; /* XXX should be an internal parameter */

  if (torus->FirstTriangle)
    DestroyObjectTriangleMesh(torus);

  for (i=0 ; i<DetailLevel ; i++)
    for (j=0 ; j<DetailLevel ; j++)
    {
      P1[0] = (ttorus->Major+ttorus->Minor*cos(j*2.0*M_PI/DetailLevel))*cos(i*2.0*M_PI/DetailLevel);
      P1[1] = ttorus->Minor*sin(j*2.0*M_PI/DetailLevel);
      P1[2] = (ttorus->Major+ttorus->Minor*cos(j*2.0*M_PI/DetailLevel))*sin(i*2.0*M_PI/DetailLevel);
      P2[0] = (ttorus->Major+ttorus->Minor*cos((j+1)*2.0*M_PI/DetailLevel))*cos(i*2.0*M_PI/DetailLevel);
      P2[1] = ttorus->Minor*sin((j+1)*2.0*M_PI/DetailLevel);
      P2[2] = (ttorus->Major+ttorus->Minor*cos((j+1)*2.0*M_PI/DetailLevel))*sin(i*2.0*M_PI/DetailLevel);
      P3[0] = (ttorus->Major+ttorus->Minor*cos(j*2.0*M_PI/DetailLevel))*cos((i+1)*2.0*M_PI/DetailLevel);
      P3[1] = ttorus->Minor*sin(j*2.0*M_PI/DetailLevel);
      P3[2] = (ttorus->Major+ttorus->Minor*cos(j*2.0*M_PI/DetailLevel))*sin((i+1)*2.0*M_PI/DetailLevel);
      P4[0] = (ttorus->Major+ttorus->Minor*cos((j+1)*2.0*M_PI/DetailLevel))*cos((i+1)*2.0*M_PI/DetailLevel);
      P4[1] = ttorus->Minor*sin((j+1)*2.0*M_PI/DetailLevel);
      P4[2] = (ttorus->Major+ttorus->Minor*cos((j+1)*2.0*M_PI/DetailLevel))*sin((i+1)*2.0*M_PI/DetailLevel);
     
      N1[0] = cos(i*2.0*M_PI/DetailLevel)*cos(j*2.0*M_PI/DetailLevel);
      N1[1] = sin(j*2.0*M_PI/DetailLevel);
      N1[2] = sin(i*2.0*M_PI/DetailLevel)*cos(j*2.0*M_PI/DetailLevel);
      N2[0] = cos(i*2.0*M_PI/DetailLevel)*cos((j+1)*2.0*M_PI/DetailLevel);
      N2[1] = sin((j+1)*2.0*M_PI/DetailLevel);
      N2[2] = sin(i*2.0*M_PI/DetailLevel)*cos((j+1)*2.0*M_PI/DetailLevel);
      N3[0] = cos((i+1)*2.0*M_PI/DetailLevel)*cos(j*2.0*M_PI/DetailLevel);
      N3[1] = sin(j*2.0*M_PI/DetailLevel);
      N3[2] = sin((i+1)*2.0*M_PI/DetailLevel)*cos(j*2.0*M_PI/DetailLevel);
      N4[0] = cos((i+1)*2.0*M_PI/DetailLevel)*cos((j+1)*2.0*M_PI/DetailLevel);
      N4[1] = sin((j+1)*2.0*M_PI/DetailLevel);
      N4[2] = sin((i+1)*2.0*M_PI/DetailLevel)*cos((j+1)*2.0*M_PI/DetailLevel);
      if (torus->Trans)
      {
        MEvaluatePoint(P1, torus->Trans, P1);
        MEvaluatePoint(P2, torus->Trans, P2);
        MEvaluatePoint(P3, torus->Trans, P3);
        MEvaluatePoint(P4, torus->Trans, P4);
        MEvaluateVector(N1, torus->Trans, N1);
        MEvaluateVector(N2, torus->Trans, N2);
        MEvaluateVector(N3, torus->Trans, N3);
        MEvaluateVector(N4, torus->Trans, N4);
      }
      AddTriangleToObjectMesh(torus, P1,P2,P3, N1,N2,N3);
      AddTriangleToObjectMesh(torus, P4,P2,P3, N4,N2,N3);
    }
}

/*****************************************************************************
*  giram_torus_inside
******************************************************************************/
static gboolean giram_torus_inside(ObjectStruct *torus, double x, double y, double z)
{ /* XXX: clipped_by */
  TorusStruct *ttorus = (TorusStruct *)torus;
  Vector vect;
  double p, pp;

  vect[0] = x;
  vect[1] = y;
  vect[2] = z;
  if (torus->Trans)
  {
    MInverseEvaluatePoint(vect, torus->Trans, vect);
  }
  p = sqrt(vect[0]*vect[0] + vect[2]*vect[2]);
  pp = vect[1]*vect[1] + (p-ttorus->Major)*(p-ttorus->Major);
  if (pp <= ttorus->Minor*ttorus->Minor)
    return !(torus->Inverse);
  else
    return torus->Inverse;
}

/*****************************************************************************
*  giram_torus_is_intersection
******************************************************************************/
static gboolean giram_torus_is_intersection(ObjectStruct *torus, Vector origin, Vector direction)
{ /* XXX: clipped_by */
  Vector org, dir;
  int i;
  double R2, r2;
  double Py2, Dy2, PDy2, k1, k2;
  double c[5], s[4];
  TorusStruct *ttorus = (TorusStruct *)torus;

  if (torus->Trans)
  {
    MInverseEvaluatePoint(org, torus->Trans, origin);
    MInverseEvaluateVector(dir, torus->Trans, direction);
  } else
  {
    for (i=0 ; i<3 ; i++)
    {
      org[i] = origin[i];
      dir[i] = direction[i];
    }
  }

  R2   = ttorus->Major * ttorus->Major;
  r2   = ttorus->Minor * ttorus->Minor;

  Py2  = org[1] * org[1];
  Dy2  = dir[1] * dir[1];
  PDy2 = org[1] * dir[1];

  k1   = org[0] * org[0] + org[2] * org[2] + Py2 - R2 - r2;
  k2   = org[0] * dir[0] + org[2] * dir[2] + PDy2;

  c[0] = 1.0;
  c[1] = 4.0 * k2;
  c[2] = 2.0 * (k1 + 2.0 * (k2 * k2 + R2 * Dy2));
  c[3] = 4.0 * (k2 * k1 + 2.0 * R2 * PDy2);
  c[4] = k1 * k1 + 4.0 * R2 * (Py2 - r2);

  return SolveQuartic(c, s);
}

/*****************************************************************************
*  giram_torus_find_intersection_segment FIXME
******************************************************************************/
static gboolean giram_torus_find_intersection_segment(ObjectStruct *torus,
                                                      Vector in_point, Vector out_point,
                                                      Vector inter_point, Vector inter_norm)
{
  return FALSE;
}

/****************************************************************************/
double Xorg, Yorg, Radius, MinorRadius;
GdkColormap *colormap;
GdkGC *gcPurple;
GdkColor Purple;
guint id;
Vector Center;

#define NONE           1
#define MOVING         2
#define BETWEEN_RADIUS 3
#define MINOR_RADIUS   4

int TORUS_STATE = NONE;

/*****************************************************************************
*  ToolInitNewTorus
******************************************************************************/
void ToolInitNewTorus(GtkWidget *DrawingArea, GdkEventButton *bevent)
{
  ViewUserDataStruct *UserData;
  GdkColormap *colormap;
  GtkWidget *StatusBar;
  double Zoom, Xoff, Yoff;

  if (TORUS_STATE == BETWEEN_RADIUS)
  {
    TORUS_STATE = MINOR_RADIUS;
    return;
  }
  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
  id = gtk_statusbar_get_context_id(GTK_STATUSBAR(StatusBar), "new torus");
  MouseToReal(bevent->x, bevent->y, DrawingArea, &Xorg, &Yorg);

  colormap = gdk_window_get_colormap(DrawingArea->window);
  gcPurple = gdk_gc_new(DrawingArea->window);
  Purple.red = 65535; Purple.green = 0; Purple.blue = 65535;
  /* XXX: the color should be settable... */
  gdk_color_alloc(colormap, &Purple);
  gdk_gc_set_foreground(gcPurple, &Purple);
  gdk_gc_set_function(gcPurple, GDK_XOR);

  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      Center[0] = Xorg;
      Center[1] = Yorg;
      Center[2] = 0.0;
      break;
    case ORTHO_XZ_CAMERA:
      Center[0] = Xorg;
      Center[1] = 0.0;
      Center[2] = Yorg;
      break;
    case ORTHO_ZY_CAMERA:
      Center[0] = 0.0;
      Center[1] = Yorg;
      Center[2] = Xorg;
      break;
  }
  Radius = 0.0;

  TORUS_STATE = MOVING;
  gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, _("Torus: "));
  gdk_pointer_grab(DrawingArea->window, FALSE,
                   GDK_BUTTON_PRESS_MASK | 
                   GDK_POINTER_MOTION_MASK | 
                   GDK_BUTTON_RELEASE_MASK,
                   NULL, NULL, bevent->time);
}

/*****************************************************************************
*  ToolMotionNewTorus
******************************************************************************/
void ToolMotionNewTorus(GtkWidget *DrawingArea, GdkEvent *event)
{
  GdkEventMotion *Mev = (GdkEventMotion *)event;
  ViewUserDataStruct *UserData;
  GtkWidget *StatusBar;
  double Zoom, Xoff, Yoff, Xold, Yold;
  gchar *Message;
  double NewRadius;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;
  if (TORUS_STATE == MOVING)
  {
    gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);

    gdk_draw_arc(DrawingArea->window, gcPurple, FALSE,
                 DrawingArea->allocation.width/2+(Xorg-Radius-Xoff)*Zoom,
                 DrawingArea->allocation.height/2-(Yorg+Radius-Yoff)*Zoom,
                 Radius*2.0*Zoom, Radius*2.0*Zoom,  0, 64*360);
    MouseToReal(Mev->x, Mev->y, DrawingArea, &Xold, &Yold);
    Radius = sqrt((Xold-Xorg)*(Xold-Xorg)+(Yold-Yorg)*(Yold-Yorg));
    gdk_draw_arc(DrawingArea->window, gcPurple, FALSE,
                 DrawingArea->allocation.width/2+(Xorg-Radius-Xoff)*Zoom,
                 DrawingArea->allocation.height/2-(Yorg+Radius-Yoff)*Zoom,
                 Radius*2.0*Zoom, Radius*2.0*Zoom,  0, 64*360);
    Message = g_strdup_printf(_("Torus: Center <%g, %g, %g>, Major Radius: %g"),
                              Center[0], Center[1], Center[2], Radius);
    gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, Message);
    g_free(Message);
  } else if ((TORUS_STATE == MINOR_RADIUS) || (TORUS_STATE == BETWEEN_RADIUS))
  {
    gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);
    gdk_draw_arc(DrawingArea->window, gcPurple, FALSE,
                 DrawingArea->allocation.width/2+(Xorg-(Radius-MinorRadius)-Xoff)*Zoom,
                 DrawingArea->allocation.height/2-(Yorg+(Radius-MinorRadius)-Yoff)*Zoom,
                 (Radius-MinorRadius)*2.0*Zoom, 
                 (Radius-MinorRadius)*2.0*Zoom, 0,64*360);
    gdk_draw_arc(DrawingArea->window, gcPurple, FALSE,
                 DrawingArea->allocation.width/2+(Xorg-(Radius+MinorRadius)-Xoff)*Zoom,
                 DrawingArea->allocation.height/2-(Yorg+(Radius+MinorRadius)-Yoff)*Zoom,
                 (Radius+MinorRadius)*2.0*Zoom, 
                 (Radius+MinorRadius)*2.0*Zoom, 0,64*360);
    MouseToReal(Mev->x, Mev->y, DrawingArea, &Xold, &Yold);
    NewRadius = sqrt((Xold-Xorg)*(Xold-Xorg)+(Yold-Yorg)*(Yold-Yorg));
    MinorRadius = fabs(NewRadius-Radius);
    gdk_draw_arc(DrawingArea->window, gcPurple, FALSE,
                 DrawingArea->allocation.width/2+(Xorg-(Radius-MinorRadius)-Xoff)*Zoom,
                 DrawingArea->allocation.height/2-(Yorg+(Radius-MinorRadius)-Yoff)*Zoom,
                 (Radius-MinorRadius)*2.0*Zoom, 
                 (Radius-MinorRadius)*2.0*Zoom, 0,64*360);
    gdk_draw_arc(DrawingArea->window, gcPurple, FALSE,
                 DrawingArea->allocation.width/2+(Xorg-(Radius+MinorRadius)-Xoff)*Zoom,
                 DrawingArea->allocation.height/2-(Yorg+(Radius+MinorRadius)-Yoff)*Zoom,
                 (Radius+MinorRadius)*2.0*Zoom, 
                 (Radius+MinorRadius)*2.0*Zoom, 0,64*360);
    Message = g_strdup_printf(_("Torus: Center <%g, %g, %g>, Major Radius: %g, Minor Radius: %g"),
            Center[0], Center[1], Center[2], Radius, MinorRadius);
    gtk_statusbar_push(GTK_STATUSBAR(StatusBar), id, Message);
    g_free(Message);
  }
}

/*****************************************************************************
*  ToolReleaseNewTorus
******************************************************************************/
void ToolReleaseNewTorus(GtkWidget *DrawingArea, GdkEvent *event)
{
  GdkEventButton *Bev = (GdkEventButton *)event;
  ViewUserDataStruct *UserData;
  GtkWidget          *StatusBar;
  double              Zoom, Xoff, Yoff;
  GSList             *tmp_list;
  FrameViewStruct    *TmpView;
  ObjectStruct       *Torus;
  TorusStruct        *TTorus;
  Vector              Vect;

  if (TORUS_STATE == MOVING)
  {
    TORUS_STATE = BETWEEN_RADIUS;
    MinorRadius = 0.0;
    return;
  }

  if (TORUS_STATE != MINOR_RADIUS)
    return;

  UserData = get_view_user_data();
  StatusBar = UserData->StatusBar;
  Zoom = UserData->ZoomFactor;
  Xoff = UserData->Xoffset;
  Yoff = UserData->Yoffset;

  gtk_statusbar_pop(GTK_STATUSBAR(StatusBar), id);

  Torus = giram_torus_new(Radius, MinorRadius);
  switch (UserData->CameraStyle)
  {
    case ORTHO_XY_CAMERA:
      Vect[0] = 90.0;
      Vect[1] = 0.0;
      Vect[2] = 0.0;
      giram_object_rotate(Torus, Vect);
      break;
    case ORTHO_ZY_CAMERA:
      Vect[0] = 0.0;
      Vect[1] = 0.0;
      Vect[2] = 90.0;
      giram_object_rotate(Torus, Vect);
      break;
  }
  giram_object_translate(Torus, Center);
  UserData->FrameView->all_objects = g_slist_append(UserData->FrameView->all_objects,
                                                    Torus);
  giram_object_build_triangle_mesh(Torus);
  TORUS_STATE = NONE;
  gdk_gc_destroy(gcPurple);
  for (tmp_list = UserData->FrameView->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, UserData->FrameView);
  gdk_pointer_ungrab(Bev->time);
}

/*****************************************************************************
*  giram_torus_new
******************************************************************************/
ObjectStruct *giram_torus_new(double major_radius, double minor_radius)
{
  ObjectStruct *torus;
  TorusStruct *ttorus;
  static GiramObjectClass *torus_klass = NULL;

  if (torus_klass == NULL)
  {
    torus_klass = giram_object_class_new();

    torus_klass->build_triangle_mesh       = giram_torus_build_triangle_mesh;
    torus_klass->inside                    = giram_torus_inside;
    torus_klass->is_intersection           = giram_torus_is_intersection;
    torus_klass->find_intersection_segment = giram_torus_find_intersection_segment;
  }
  ttorus = g_new(TorusStruct, 1);
  torus = (ObjectStruct *)ttorus;
  InitObject(torus);
  torus->Type = TORUS_OBJECT;
  torus->klass = torus_klass;
  ttorus->Major = major_radius;
  ttorus->Minor = minor_radius;
  return torus;
}

/*****************************************************************************
*  PDBCreateTorus
******************************************************************************/
int PDBCreateTorus(int *Id, double *MajorRadius, double *MinorRadius)
{
  ObjectStruct    *Torus;
  TorusStruct     *TTorus;
  FrameStruct     *LocalFrame;
  GSList          *tmp_list;
  FrameViewStruct *TmpView;

  Torus = giram_torus_new(*MajorRadius, *MinorRadius);

  for (LocalFrame = FirstFrame ;
       LocalFrame->Id != *Id ;
       LocalFrame = LocalFrame->Next)
                  ;
  LocalFrame->all_objects = g_slist_append(LocalFrame->all_objects, Torus);
  giram_object_build_triangle_mesh(Torus);
  for (tmp_list = LocalFrame->all_views ;
       tmp_list ;
       tmp_list = tmp_list->next)
  {
    TmpView = tmp_list->data;
    gtk_widget_queue_draw(TmpView->drawing_area);
  }
  RebuildCSGTree(NULL, LocalFrame);
  return Torus->Id;
}

